import enum
from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class Mode(enum.IntEnum):
    PLAY = 0
    WATCH = 1


class Autoencode(POPGymEnv):
    """A game where the agent must press buttons in order it saw
    them pressed. E.g., seeing [1, 2, 3] means I should press them in the order
    [1, 2, 3].

    Args:
        num_decks: The maximum number of decks the agent must memorize

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1):
        self.deck = Deck(num_decks)
        self.deck.add_players("system")
        self.max_episode_length = self.deck.num_cards * 2 - 1
        self.action_space = self.deck.get_obs_space(["suits"])
        self.observation_space = gym.spaces.Tuple(
            (
                gym.spaces.Discrete(2),
                self.action_space,
            )
        )
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete([4] * self.deck.num_cards),
                gym.spaces.Discrete(2),
                gym.spaces.Box(0, 1, (1,)),
            )
        )
        self.mode = Mode.WATCH

    def make_obs(self, card_idx):
        card_suit = self.deck.suits_idx[card_idx].reshape(-1)
        return int(self.mode.value), card_suit.item()

    def get_state(self):
        cards = self.deck.suits_idx[self.deck.idx].copy()
        mode = int(self.mode.value)
        pos = np.array(
            [len(self.deck["system"]) / self.deck.num_cards], dtype=np.float32
        )
        return cards, mode, pos

    def step(self, action):
        done = False
        reward = 0
        # TODO: This is causing flaky tests, make sure we are not
        # off-by-one in step
        reward_scale = 1 / self.deck.num_cards

        if self.mode == Mode.WATCH:
            self.deck.deal("system", 1)
            if len(self.deck) == 0:
                self.mode = Mode.PLAY
                # Flip the cards so they play backwards
                # self.deck["system"].reverse()
            obs = self.make_obs(self.deck["system"][-1])
        else:
            # Recited all cards
            done = len(self.deck["system"]) == 1
            correct_card = self.deck.suits_idx[self.deck["system"].pop(-1)]
            if action == correct_card:
                reward = reward_scale
            else:
                reward = -reward_scale

            obs = self.make_obs(np.array([0]))

        info = {}

        return obs, reward, done, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("system", 1)

        self.mode = Mode.WATCH
        obs = self.make_obs(self.deck["system"][-1])
        if return_info:
            return obs, {}

        return obs


class AutoencodeEasy(Autoencode):
    pass


class AutoencodeMedium(Autoencode):
    def __init__(self):
        super().__init__(num_decks=2)


class AutoencodeHard(Autoencode):
    def __init__(self):
        super().__init__(num_decks=3)



from typing import List, Tuple

import gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class Frameconv(BaseModel):
    r"""Temporal convolution over a series of observations. stack_size determines
    the temporal extent of the filter. See

    @article{bai_empirical_2018,
        title = {
            An empirical evaluation of generic convolutional and recurrent networks
            for sequence modeling
        },
        journal = {arXiv preprint arXiv:1803.01271},
        author = {Bai, Shaojie and Kolter, J Zico and Koltun, Vladlen},
        year = {2018},
    }
    """

    MODEL_CONFIG = {
        # Number of consecutive frames to stack into a single input
        "stack_size": 4,
        "embedding": "sine",
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.core = nn.Conv1d(
            self.cfg["preprocessor_output_size"],
            self.cfg["hidden_size"],
            self.cfg["stack_size"],
            1,
        )

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(
                self.cfg["stack_size"] - 1, self.cfg["preprocessor_output_size"]
            ),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        assert self.cfg["hidden_size"] % self.cfg["stack_size"] == 0, (
            f"Hidden size {self.cfg['hidden_size']} must be divisible by stack size"
            f" {self.cfg['stack_size']}"
        )
        B, T, F = z.shape
        # [B, T + stack_size - 1, F]
        time_padded = torch.cat([state[0], z], dim=1)

        out = self.core(time_padded.permute(0, 2, 1)).permute(0, 2, 1)
        state = time_padded[:, -self.cfg["stack_size"] + 1 :]

        # State expected to be list
        return out, [state]



import enum
from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.env import POPGymEnv


class HiddenSquare(enum.IntEnum):
    CLEAR = 0
    MINE = 1
    VIEWED = 2


class MineSweeper(POPGymEnv):
    """Classic Microsoft MineSweeper but with the board obscured. Mines are hidden
    underneath a grid of tiles. The player clicks a tile, which returns the
    coordinates of the tile and how many mines are present in adjacent tiles.
    Clicking a mine loses. Player must click all free squares to win.

    Args:
        difficulty: Easy, medium, or hard. Sets the board size and number of
            mines based on difficulty.

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, difficulty="easy"):
        assert difficulty in ["easy", "medium", "hard"]
        if difficulty == "easy":
            dims = 4, 4
            num_mines = 2
        elif difficulty == "medium":
            dims = 6, 6
            num_mines = 6
        elif difficulty == "hard":
            dims = 8, 8
            num_mines = 10
        else:
            raise NotImplementedError(f"Invalid difficulty {difficulty}")

        self.dims = dims
        self.num_mines = num_mines
        self.max_episode_length = dims[0] * dims[1] - num_mines
        self.success_reward_scale = 1 / self.max_episode_length
        self.fail_reward_scale = -0.5 - self.success_reward_scale
        # -1 for one step less (last action must be mine for lowest G)
        # -1 for one step less (first action is a view which will give reward)
        self.bad_action_reward_scale = -0.5 / (self.max_episode_length - 2)
        self.observation_space = gym.spaces.Discrete(min(num_mines + 1, 10))
        self.state_space = gym.spaces.MultiDiscrete([3] * np.prod(self.dims))
        self.action_space = gym.spaces.MultiDiscrete(np.array(dims))

    def get_state(self):
        return self.hidden_grid.flatten().copy()

    def step(self, action):
        done = False
        action = tuple(action)
        if self.hidden_grid[action] == HiddenSquare.MINE:
            done = True
            reward = self.fail_reward_scale
        elif self.hidden_grid[action] == HiddenSquare.VIEWED:
            # Querying already viewed square
            reward = self.bad_action_reward_scale
        else:
            self.hidden_grid[action] = HiddenSquare.VIEWED
            reward = self.success_reward_scale

        done |= (self.timestep == self.max_episode_length) or np.all(
            self.hidden_grid != HiddenSquare.CLEAR
        ).item()

        obs = self.neighbor_grid[action].item()
        self.timestep += 1

        return obs, reward, done, {}

    def render(self):
        visible_mask = self.hidden_grid == HiddenSquare.VIEWED
        result = np.full(self.hidden_grid.shape, ".")
        result[visible_mask] = self.neighbor_grid[visible_mask]
        out = " " + str(result).replace("[", "").replace("]", "").replace("'", "")
        print(out)

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        # Init grids
        self.hidden_grid = np.full(self.dims, HiddenSquare.CLEAR, dtype=np.int8)
        mines_flat = self.np_random.choice(
            np.arange(self.dims[0] * self.dims[1]), size=self.num_mines, replace=False
        )
        self.hidden_grid.ravel()[mines_flat] = HiddenSquare.MINE
        self.neighbor_grid = np.zeros_like(self.hidden_grid)
        src_grid = np.pad(self.hidden_grid, [(1, 1), (1, 1)], constant_values=0)
        for shift_i in [-1, 0, 1]:
            for shift_j in [-1, 0, 1]:
                self.neighbor_grid += np.roll(
                    np.roll(src_grid, shift_i, 0), shift_j, 1
                )[1:-1, 1:-1]

        self.timestep = 0
        obs = 0

        if return_info:
            return obs, {}

        return obs


if __name__ == "__main__":
    e = MineSweeper()
    obs = e.reset()
    e.render()
    done = False
    while not done:
        action = np.array(input("Enter x,y:").split(",")).astype(np.int8)
        obs, reward, done, info = e.step(action)
        e.render()
        print("reward", reward)


class MineSweeperEasy(MineSweeper):
    def __init__(self):
        super().__init__("easy")


class MineSweeperMedium(MineSweeper):
    def __init__(self):
        super().__init__("medium")


class MineSweeperHard(MineSweeper):
    def __init__(self):
        super().__init__("hard")



from gym import spaces

from popgym.core.env import POPGymEnv
from popgym.core.observability import OBS, STATE, Observability
from popgym.core.wrapper import POPGymWrapper


class Markovian(POPGymWrapper):
    """Wrapper that adds the hidden Markov state to the observation or info dict

    Args:
        env: The environment.
        observability: The observability level.

    Returns:
        A gym environment
    """

    def __init__(self, env: POPGymEnv, observability: Observability):
        super(Markovian, self).__init__(env)
        assert isinstance(
            env.unwrapped, POPGymEnv
        ), "This wrapper is made for POPGymEnvs."
        self.observability = observability
        if observability == Observability.FULL_AND_PARTIAL:
            self.observation_space = spaces.Dict(
                {
                    OBS: self.env.observation_space,
                    STATE: self.state_space,
                }
            )
        elif observability == Observability.FULL:
            self.observation_space = self.state_space
        elif observability in [Observability.PARTIAL, Observability.FULL_IN_INFO_DICT]:
            pass
        else:
            raise NotImplementedError("Invalid observability level:", observability)

    def reset(self, **kwargs):
        if kwargs.get("return_info", False):
            obs, info = self.env.reset(**kwargs)
            obs, info = self.add_state(obs, info)
            return obs, info
        else:
            obs = self.env.reset(**kwargs)
            obs, _ = self.add_state(obs, {})
            return obs

    def add_state(self, obs, info):
        if self.observability == Observability.FULL:
            obs = self.get_state()
        elif self.observability == Observability.FULL_AND_PARTIAL:
            obs = {OBS: obs, STATE: self.get_state()}
        elif self.observability == Observability.FULL_IN_INFO_DICT:
            info[STATE] = self.get_state()
        return obs, info

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        obs, info = self.add_state(obs, info)
        return obs, reward, done, info



import math

import torch
from torch import nn


def get_aggregator(name: str) -> nn.Module:
    assert name in [
        "sum",
        "max",
    ], "Invalid aggregator. Must be 'sum' or 'max'"
    return {
        "sum": SumAggregation,
        "max": MaxAggregation,
    }[name]


class Aggregation(nn.Module):
    """Aggregates (x_k ... x_t , s_k) into s_t"""

    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(
        self,
        x: torch.Tensor,
        memory: torch.Tensor,
    ) -> torch.Tensor:
        raise NotImplementedError()


class SumAggregation(Aggregation):
    def forward(
        self,
        x: torch.Tensor,
        memory: torch.Tensor,
    ) -> torch.Tensor:
        return x.cumsum(dim=1).clamp(-1e20, 1e20) + memory


class MaxAggregation(Aggregation):
    def forward(
        self,
        x: torch.Tensor,
        memory: torch.Tensor,
    ) -> torch.Tensor:
        return torch.maximum(x.cummax(dim=1).values, memory)



import unittest

from popgym.envs.repeat_first import RepeatFirst
from tests.base_env_test import AbstractTest


class TestRepeatFirst(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = RepeatFirst()

    def test_perfect(self):
        done = False
        init_item = self.env.reset()
        reward = 0
        for i in range(51):
            self.assertFalse(done)
            item, rew, done, info = self.env.step(init_item)
            self.assertTrue(item < 4)
            reward += rew

        self.assertTrue(done)
        self.assertAlmostEqual(1.0, reward)



import math

import torch
from torch import nn


def positional_encoding(max_len, d_model):
    pe = torch.zeros(max_len, d_model)
    position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
    # Changed log from 10_000.0 to max_len, improves accuracy for hard labyrinth
    div_term = torch.exp(
        torch.arange(0, d_model, 2).float() * (-math.log(max_len) / d_model)
    )
    pe[:, 0::2] = torch.sin(position * div_term)
    pe[:, 1::2] = torch.cos(position * div_term)
    return pe


@torch.jit.script
def _psi(ab: torch.Tensor, c: torch.Tensor, t_minus_i: torch.Tensor) -> torch.Tensor:
    # ab: [1, T, F ... F]
    # c: [1, T, 1 ... 1]
    # t_minus_i: [1, T, 1 ... 1]
    return torch.exp(c * ab * t_minus_i)


@torch.jit.script
def _compute_state(
    ab: torch.Tensor,
    c: torch.Tensor,
    x: torch.Tensor,
    n_to_zero: torch.Tensor,
    one_to_np1: torch.Tensor,
    T: int,
    prev_state: torch.Tensor,
):
    state = torch.cumsum(x * _psi(ab, c, n_to_zero[:T]), dim=1) * _psi(
        ab, c, -n_to_zero[:T]
    )
    # Add to previous output
    # Shift the previous inputs by 1 to T
    # because they are now 1 to T timesteps older with the
    # addition of T new timesteps present in x
    psi_shift = _psi(ab, c, one_to_np1[:T])
    shifted_prev_state = prev_state * psi_shift
    output = state + shifted_prev_state
    return output


class PhaserEncoding(nn.Module):
    def __init__(
        self,
        max_len: int,
        d_model: int,
        num_dims: int,
        max_period: int = 1024,
        dtype: torch.dtype = torch.double,
        decay_bias: float = -0.5,
        freq_scale: float = 1.0,
        fudge_factor: float = 0.025,
        soft_clamp: bool = False,
    ):
        """A phaser-encoded aggregation operator
        Inputs:
            max_len: Maximum length of the batch in timesteps. Note this
                is not the episode length, but rather the batch length.
            d_model: Feature dimension of the model
            num_dims: Number of feature dimensions of the model, each should be
                of size d_model
            max_period: This determines the initial maximum sinusoidal period. It should
                be set to the maximum episode length if known. Until the parameters
                change, the model will be unable to differentiate relative time
                differences greater than max_period.
            dtype: Whether to use floats or doubles. Note doubles enables
                significantly more representational power for a little
                extra compute
            decay_bias: How much bias to exponential decay. Higher -> biased
                towards shorter-term memory. This should roughly be between 0 and 5
                since this value goes through a sigmoid activation.
            freq_scale: How much to bias the sinusoidal part of the encoding.
                Lower -> lower frequencies, biased towards longer-term memory
                    and being unable to tell the difference between x_t and x_{t+1}
                Higher -> higher frequencies, biased towards quick reactions/control
                    and more likely to differentiate between x_t and x_{t+1}
                This should be a positive number.
            fudge_factor: A small positive number to prevent overflows
        """
        super().__init__()
        self.num_dims = num_dims
        self.d_model = d_model
        # Set the upper bound to zero because our prior knowledge
        # dictates experiences from the past never become stronger,
        # only fade. Also, this could cause explosions as we magnify
        # values at each timestep
        # To prevent overflows, ensure exp(limit * max_len) < {float,double}
        # limit * max_len < log({float,double})
        # limit == log({float,double}) / max_len - fudge_factor
        assert dtype in [torch.float, torch.double]
        self.dtype = dtype
        dtype_max = torch.finfo(dtype).max
        self.limit = math.log(dtype_max) / max_len - fudge_factor
        self.soft_clamp = soft_clamp

        # [1, 1 | 1, ... 1, F]
        # [B, T | F, ... F, F]
        param_shape = [1, 1] + [self.d_model] * num_dims
        a = torch.linspace(
            0 - fudge_factor, -self.limit + fudge_factor, self.d_model**num_dims
        ).reshape(param_shape)
        b = (
            2
            * torch.pi
            / torch.linspace(max_period, torch.pi, self.d_model**num_dims).reshape(
                param_shape
            )
        )
        # b = b.flatten()[torch.randperm(b.numel())].reshape(b.shape)
        self.c = nn.Parameter(torch.tensor(decay_bias))
        self.d = nn.Parameter(torch.tensor(freq_scale))
        self.ab = nn.Parameter(torch.complex(a, b))

        torch.nn.init.normal_(self.ab.real, mean=0, std=0.01)
        self.ab.real.data.clamp_(-self.limit + fudge_factor, 0 - fudge_factor)

        n_to_zero = torch.arange(max_len, dtype=dtype).flip(0)
        self.register_buffer("n_to_zero", n_to_zero)
        one_to_np1 = torch.arange(1, max_len + 1, dtype=dtype)
        self.register_buffer("one_to_np1", one_to_np1)

    def psi(self, t_minus_i):
        assert t_minus_i.dim() == 1

        T = t_minus_i.shape[0]
        F = [self.d_model] * self.num_dims
        broadcast = [1] * len(F)
        # Compute for all filters/fourier series terms
        # e^(t * (a + bi))
        if self.soft_clamp:
            real = -self.limit * (self.c + self.ab.real).sigmoid()
        else:
            real = (self.c + self.ab.real).clamp(-self.limit, 0)
        imag = self.d * self.ab.imag
        return torch.exp(
            torch.complex(real, imag) * t_minus_i.reshape(1, T, *broadcast)
        )
        # return _psi(self.ab, self.c, t_minus_i.reshape(1, T, *broadcast))

    def recurrent_update2(self, x, state):
        B, T, *F = x.shape

        prev_state = torch.view_as_complex(state)
        ones = torch.tensor([-1.0, 1.0], dtype=torch.complex128, device=x.device)

        if self.training:
            psi_minus_one, psi_one = self.psi(ones).unbind(1)
        else:
            psi_minus_one, psi_one = self.psi(ones).unbind(1)
        psi_n_to_zero = psi_one.expand(1, T, -1, -1, -1).cumprod(1).flip(1)
        minus_psi_n_to_zero = psi_minus_one.expand(1, T, -1, -1, -1).cumprod(1).flip(1)

        # Convert back to float for performance
        state = torch.cumsum(x * psi_n_to_zero, dim=1) * minus_psi_n_to_zero

        shift = psi_n_to_zero.flip(1) * psi_one
        shifted_prev_state = prev_state * shift
        output = state + shifted_prev_state
        return output.to(torch.complex64)

    def recurrent_update(self, x, state):
        B, T, *F = x.shape
        prev_state = torch.view_as_complex(state)

        state = torch.cumsum(x * self.psi(self.n_to_zero[:T]), dim=1) * self.psi(
            -self.n_to_zero[:T]
        )
        # Add to previous output
        # Shift the previous inputs by 1 to T
        # because they are now 1 to T timesteps older with the
        # addition of T new timesteps present in x
        psi_shift = self.psi(self.one_to_np1[:T])
        shifted_prev_state = prev_state * psi_shift
        output = state + shifted_prev_state
        return output.to(torch.complex64)

    def forward(self, x, state):
        return self.recurrent_update(x, state)


class ScaledPositionalEncoding(nn.Embedding):
    def __init__(self, num_embeddings, embedding_dim, **kwargs):
        super().__init__(num_embeddings, embedding_dim, **kwargs)
        self.a = nn.Parameter(torch.rand(embedding_dim))
        num_zero_weights = round(0.75 * embedding_dim)
        zero_idx = torch.randperm(embedding_dim)[:num_zero_weights]
        self.a.data[zero_idx] = 0

    def forward(self, idx):
        return super().forward(idx) * self.a


class SimpleScaledEncoding(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()
        v = torch.zeros(d_model)
        v[-4:] = 1.0 / max_len
        self.register_buffer("v", v)

    def forward(self, idx):
        return idx.reshape(-1, 1) * self.v


class Time2Vec(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()
        self.d_model = d_model
        self.max_len = max_len
        self.net = nn.Linear(1, d_model)
        # self.emb = nn.Embedding(max_len, 2 * d_model)

    def forward(self, idx):
        # [B * T, 1]
        idx = idx.reshape(-1, 1)
        emb = self.net(idx.float()).reshape(-1, self.d_model)
        zero, rest = emb.split([1, self.d_model - 1], dim=-1)
        return torch.cat([zero, rest.sin()], dim=-1)


class Base2Embedding(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()
        self.d_model = d_model
        self.max_len = max_len
        mask = 2 ** torch.arange(d_model)
        # Overflow, zero these for sanity
        # seq_len <= 65536
        mask[16:] = mask[16]
        assert max_len < 2**16 - 1
        self.register_buffer("mask", mask)

    def forward(self, idx):
        return idx.unsqueeze(-1).bitwise_and(self.mask).ne(0).float()


class PerceptronEmbedding(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()
        self.d_model = d_model
        self.lin = nn.Linear(1, d_model, bias=False)

    def forward(self, idx):
        idx = idx.float()
        return self.lin(idx.reshape(-1, 1)).sin()


class SoftExponentialEmbedding(nn.Module):
    def __init__(self, max_len, d_model):
        super().__init__()
        self.d_model = d_model
        self.a = nn.Parameter([-1.0])
        self.lin = nn.Linear(1, d_model, bias=False)

    def forward(self, idx):
        idx = idx.float()
        self.a.data = self.a.data.clamp_(-1.0, 1e-4)
        (torch.exp(self.a * idx) - 1) / self.a + self.a
        return self.lin(idx.reshape(-1, 1)).sin()



from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.envs.multiarmed_bandit import MultiarmedBandit


class NonstationaryBandit(MultiarmedBandit):
    def __init__(self, *args, **kwargs):
        super(NonstationaryBandit, self).__init__(*args, *kwargs)
        self.state_space = gym.spaces.Tuple(self.state_space, self.state_space)
        self.final_bandits = np.zeros((self.num_bandits,))

    def get_state(self):
        state = super(NonstationaryBandit, self).get_state()
        return state, self.final_bandits.copy()

    def step(self, action, increment=True):
        obs, reward, done, info = super().step(action, increment)
        self.bandits = (self.final_bandits - self.bandits) / self.max_episode_length
        return obs, reward, done, self.info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:

        out = super().reset(seed=seed, return_info=return_info, options=options)
        self.final_bandits = self.np_random.rand(self.num_bandits)
        if return_info:
            obs, info = out
        else:
            obs = out
            info = {}

        info["final_bandits"] = self.final_bandits

        if return_info:
            return obs, info
        return obs


class NonstationaryBanditEasy(NonstationaryBandit):
    def __init__(self):
        super().__init__(num_bandits=10, episode_length=200)


class NonstationaryBanditMedium(NonstationaryBandit):
    def __init__(self):
        super().__init__(num_bandits=20, episode_length=400)


class NonstationaryBanditHard(NonstationaryBandit):
    def __init__(self):
        super().__init__(num_bandits=30, episode_length=600)



from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.maze import Actions, Cell, Explored, MazeEnv


class LabyrinthExplore(MazeEnv):
    """A maze environment where the agent receives reward for visiting
    new grid squares

    Args:
        maze_dims: (heigh, width) of the generated mazes in blocks
        episode_length: maximum length of an episode

    Returns:
        A gym environment

    """

    def __init__(self, maze_dims=(10, 10), episode_length=1024):
        super().__init__(maze_dims, episode_length)
        self.neg_reward_scale = -1 / self.max_episode_length

    def step(self, action):
        new_square = self.explored[tuple(self.move(action))] == Explored.NO
        super().step(action)
        reward = self.neg_reward_scale
        done = False
        y, x = self.position
        if new_square:
            reward = self.pos_reward_scale
        if self.curr_step == self.max_episode_length - 1:
            done = True
        free_mask = self.maze.grid != Cell.OBSTACLE
        visit_mask = self.explored == Explored.YES
        if np.all(free_mask == visit_mask):
            # Explored as much as possible
            done = True

        obs = self.get_obs(action)
        info = {"position": (x, y)}

        return obs, reward, done, info

    def render(self):
        print(self.tostring(start=True, end=False, agent=True, visited=True))

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        # Based on free space
        self.pos_reward_scale = 1 / ((self.maze.grid == Cell.FREE).sum())
        y, x = self.position
        obs = self.get_obs(Actions.NONE)

        if return_info:
            return obs, {"maze": str(self.maze.grid)}
        return obs


class LabyrinthExploreEasy(LabyrinthExplore):
    def __init__(self):
        super().__init__(maze_dims=(10, 10), episode_length=1024)


class LabyrinthExploreMedium(LabyrinthExplore):
    def __init__(self):
        super().__init__(maze_dims=(14, 14), episode_length=1024)


class LabyrinthExploreHard(LabyrinthExplore):
    def __init__(self):
        super().__init__(maze_dims=(18, 18), episode_length=1024)



from tests.base_env_test import AbstractTest


from popgym.envs.multiarmed_bandit import MultiarmedBandit


class TestBandits(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = MultiarmedBandit()



# Adapted from https://github.com/HazyResearch/state-spaces
# flake8: noqa
""" Standalone version of Structured (Sequence) State Space (S4) model. """

import logging
import math
from functools import partial

import numpy as np
import opt_einsum as oe
import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange, repeat
from scipy import special as ss

contract = oe.contract
contract_expression = oe.contract_expression


def get_logger(name=__name__, level=logging.INFO) -> logging.Logger:
    """Initializes multi-GPU-friendly python logger."""

    logger = logging.getLogger(name)
    logger.setLevel(level)

    return logger


log = get_logger(__name__)

""" Cauchy and Vandermonde kernels """

try:  # Try CUDA extension
    from extensions.cauchy.cauchy import cauchy_mult

    has_cauchy_extension = True
except:
    log.warning(
        "CUDA extension for cauchy multiplication not found. Install by going to"
        " extensions/cauchy/ and running `python setup.py install`. This should speed"
        " up end-to-end training by 10-50%"
    )
    has_cauchy_extension = False

try:  # Try pykeops
    import pykeops
    from pykeops.torch import Genred

    has_pykeops = True
    log.info("Pykeops installation found.")

    def _broadcast_dims(*tensors):
        max_dim = max([len(tensor.shape) for tensor in tensors])
        tensors = [
            tensor.view((1,) * (max_dim - len(tensor.shape)) + tensor.shape)
            for tensor in tensors
        ]
        return tensors

    def cauchy_conj(v, z, w):
        """Pykeops version"""
        expr_num = "z * ComplexReal(v) - Real2Complex(Sum(v * w))"
        expr_denom = "ComplexMult(z-w, z-Conj(w))"

        cauchy_mult = Genred(
            f"ComplexDivide({expr_num}, {expr_denom})",
            [
                "v = Vj(2)",
                "z = Vi(2)",
                "w = Vj(2)",
            ],
            reduction_op="Sum",
            axis=1,
        )

        v, z, w = _broadcast_dims(v, z, w)
        v = _c2r(v)
        z = _c2r(z)
        w = _c2r(w)

        r = 2 * cauchy_mult(v, z, w, backend="GPU")
        return _r2c(r)

    def log_vandermonde(v, x, L):
        expr = "ComplexMult(v, ComplexExp(ComplexMult(x, l)))"
        vandermonde_mult = Genred(
            expr,
            [
                "v = Vj(2)",
                "x = Vj(2)",
                "l = Vi(2)",
            ],
            reduction_op="Sum",
            axis=1,
        )

        l = torch.arange(L).to(x)
        v, x, l = _broadcast_dims(v, x, l)
        v = _c2r(v)
        x = _c2r(x)
        l = _c2r(l)

        r = vandermonde_mult(v, x, l, backend="GPU")
        return 2 * _r2c(r).real

    def log_vandermonde_transpose(u, v, x, L):
        """
        u: ... H L
        v: ... H N
        x: ... H N
        Returns: ... H N

        V = Vandermonde(a, L) : (H N L)
        contract_L(V * u * v)
        """
        expr = "ComplexMult(ComplexMult(v, u), ComplexExp(ComplexMult(x, l)))"
        vandermonde_mult = Genred(
            expr,
            [
                "u = Vj(2)",
                "v = Vi(2)",
                "x = Vi(2)",
                "l = Vj(2)",
            ],
            reduction_op="Sum",
            axis=1,
        )

        l = torch.arange(L).to(x)
        u, v, x, l = _broadcast_dims(u, v, x, l)
        u = _c2r(u)
        v = _c2r(v)
        x = _c2r(x)
        l = _c2r(l)

        r = vandermonde_mult(u, v, x, l, backend="GPU")
        return _r2c(r)

except ImportError:
    has_pykeops = False
    if not has_cauchy_extension:
        log.warning(
            "Falling back on slow Cauchy kernel. Install at least one of pykeops or the"
            " CUDA extension for efficiency."
        )

        def cauchy_naive(v, z, w):
            """
            v, w: (..., N)
            z: (..., L)
            returns: (..., L)
            """
            cauchy_matrix = v.unsqueeze(-1) / (
                z.unsqueeze(-2) - w.unsqueeze(-1)
            )  # (... N L)
            return torch.sum(cauchy_matrix, dim=-2)

    # Vandermonde functions
    log.warning(
        "Falling back on slow Vandermonde kernel. Install pykeops for improved memory"
        " efficiency."
    )

    def log_vandermonde(v, x, L):
        """
        v: (..., N)
        x: (..., N)
        returns: (..., L) \sum v x^l
        """
        vandermonde_matrix = torch.exp(
            x.unsqueeze(-1) * torch.arange(L).to(x)
        )  # (... N L)
        vandermonde_prod = contract(
            "... n, ... n l -> ... l", v, vandermonde_matrix
        )  # (... L)
        return 2 * vandermonde_prod.real

    def log_vandermonde_transpose(u, v, x, L):
        vandermonde_matrix = torch.exp(
            x.unsqueeze(-1) * torch.arange(L).to(x)
        )  # (... N L)
        vandermonde_prod = contract(
            "... l, ... n, ... n l -> ... n", u.to(x), v.to(x), vandermonde_matrix
        )  # (... L)
        return vandermonde_prod


_conj = lambda x: torch.cat([x, x.conj()], dim=-1)
_c2r = torch.view_as_real
_r2c = torch.view_as_complex
if tuple(map(int, torch.__version__.split(".")[:2])) >= (1, 10):
    _resolve_conj = lambda x: x.conj().resolve_conj()
else:
    _resolve_conj = lambda x: x.conj()


""" Simple nn.Module components """


def Activation(activation=None, dim=-1):
    if activation in [None, "id", "identity", "linear"]:
        return nn.Identity()
    elif activation == "tanh":
        return nn.Tanh()
    elif activation == "relu":
        return nn.ReLU()
    elif activation == "gelu":
        return nn.GELU()
    elif activation in ["swish", "silu"]:
        return nn.SiLU()
    elif activation == "glu":
        return nn.GLU(dim=dim)
    elif activation == "sigmoid":
        return nn.Sigmoid()
    else:
        raise NotImplementedError(
            "hidden activation '{}' is not implemented".format(activation)
        )


def LinearActivation(
    d_input,
    d_output,
    bias=True,
    transposed=False,
    activation=None,
    activate=False,  # Apply activation as part of this module
    **kwargs,
):
    """Returns a linear nn.Module with control over axes order, initialization, and activation"""

    # Construct core module
    linear_cls = partial(nn.Conv1d, kernel_size=1) if transposed else nn.Linear
    if activation == "glu":
        d_output *= 2
    linear = linear_cls(d_input, d_output, bias=bias, **kwargs)

    if activate and activation is not None:
        activation = Activation(activation, dim=-2 if transposed else -1)
        linear = nn.Sequential(linear, activation)
    return linear


class DropoutNd(nn.Module):
    def __init__(self, p: float = 0.5, tie=True, transposed=True):
        """
        tie: tie dropout mask across sequence lengths (Dropout1d/2d/3d)
        """
        super().__init__()
        if p < 0 or p >= 1:
            raise ValueError(
                "dropout probability has to be in [0, 1), but got {}".format(p)
            )
        self.p = p
        self.tie = tie
        self.transposed = transposed
        self.binomial = torch.distributions.binomial.Binomial(probs=1 - self.p)

    def forward(self, X):
        """X: (batch, dim, lengths...)"""
        if self.training:
            if not self.transposed:
                X = rearrange(X, "b d ... -> b ... d")
            mask_shape = X.shape[:2] + (1,) * (X.ndim - 2) if self.tie else X.shape
            mask = torch.rand(*mask_shape, device=X.device) < 1.0 - self.p
            X = X * mask * (1.0 / (1 - self.p))
            if not self.transposed:
                X = rearrange(X, "b ... d -> b d ...")
            return X
        return X


""" Misc functional utilities """


def power(L, A, v=None):
    """Compute A^L and the scan sum_i A^i v_i

    A: (..., N, N)
    v: (..., N, L)
    """

    I = torch.eye(A.shape[-1]).to(A)  # , dtype=A.dtype, device=A.device)

    powers = [A]
    l = 1
    while True:
        if L % 2 == 1:
            I = powers[-1] @ I
        L //= 2
        if L == 0:
            break
        l *= 2
        powers.append(powers[-1] @ powers[-1])

    if v is None:
        return I

    # Invariants:
    # powers[-1] := A^l
    # l := largest po2 at most L

    # Note that an alternative divide and conquer to compute the reduction is possible and can be embedded into the above loop without caching intermediate powers of A
    # We do this reverse divide-and-conquer for efficiency reasons:
    # 1) it involves fewer padding steps for non-po2 L
    # 2) it involves more contiguous arrays

    # Take care of edge case for non-po2 arrays
    # Note that this initial step is a no-op for the case of power of 2 (l == L)
    k = v.size(-1) - l
    v_ = powers.pop() @ v[..., l:]
    v = v[..., :l]
    v[..., :k] = v[..., :k] + v_

    # Handle reduction for power of 2
    while v.size(-1) > 1:
        v = rearrange(v, "... (z l) -> ... z l", z=2)
        v = v[..., 0, :] + powers.pop() @ v[..., 1, :]
    return I, v.squeeze(-1)


""" HiPPO utilities """


def transition(measure, N):
    """A, B transition matrices for different measures"""
    # Legendre (translated)
    if measure == "legt":
        Q = np.arange(N, dtype=np.float64)
        R = (2 * Q + 1) ** 0.5
        j, i = np.meshgrid(Q, Q)
        A = R[:, None] * np.where(i < j, (-1.0) ** (i - j), 1) * R[None, :]
        B = R[:, None]
        A = -A

        # Halve again for timescale correctness
        A *= 0.5
        B *= 0.5
    # Legendre (scaled)
    elif measure == "legs":
        q = np.arange(N, dtype=np.float64)
        col, row = np.meshgrid(q, q)
        r = 2 * q + 1
        M = -(np.where(row >= col, r, 0) - np.diag(q))
        T = np.sqrt(np.diag(2 * q + 1))
        A = T @ M @ np.linalg.inv(T)
        B = np.diag(T)[:, None]
        B = (
            B.copy()
        )  # Otherwise "UserWarning: given NumPY array is not writeable..." after torch.as_tensor(B)
    elif measure == "legsd":
        # Essentially equivalent to S4D-LegS
        q = np.arange(N, dtype=np.float64)
        col, row = np.meshgrid(q, q)
        r = 2 * q + 1
        M = -(np.where(row >= col, r, 0) - np.diag(q))
        T = np.sqrt(np.diag(2 * q + 1))
        A = T @ M @ np.linalg.inv(T)
        B = np.diag(T)[:, None]
        B = (
            B.copy()
        )  # Otherwise "UserWarning: given NumPY array is not writeable..." after torch.as_tensor(B)
        A += 0.5 * B * B[None, :, 0]
        B = B / 2.0
    elif measure in ["fourier_diag", "foud"]:
        # Essentially equivalent to S4D-Lin
        freqs = np.arange(N // 2)
        d = np.stack([freqs, np.zeros(N // 2)], axis=-1).reshape(-1)[:-1]
        A = 2 * np.pi * (-np.diag(d, 1) + np.diag(d, -1))
        A = A - 0.5 * np.eye(N)
        B = np.zeros(N)
        B[0::2] = 2**0.5
        B[0] = 1
        B = B[:, None]
    elif measure in ["fourier", "fout"]:
        freqs = np.arange(N // 2)
        d = np.stack([np.zeros(N // 2), freqs], axis=-1).reshape(-1)[1:]
        A = np.pi * (-np.diag(d, 1) + np.diag(d, -1))
        B = np.zeros(N)
        B[0::2] = 2**0.5
        B[0] = 1

        # Subtract off rank correction - this corresponds to the other endpoint u(t-1) in this case
        A = A - B[:, None] * B[None, :]
        B = B[:, None]
    else:
        raise NotImplementedError

    return A, B


def rank_correction(measure, N, rank=1, dtype=torch.float):
    """Return low-rank matrix L such that A + L is normal"""

    if measure == "legs":
        assert rank >= 1
        P = torch.sqrt(0.5 + torch.arange(N, dtype=dtype)).unsqueeze(0)  # (1 N)
    elif measure == "legt":
        assert rank >= 2
        P = torch.sqrt(1 + 2 * torch.arange(N, dtype=dtype))  # (N)
        P0 = P.clone()
        P0[0::2] = 0.0
        P1 = P.clone()
        P1[1::2] = 0.0
        P = torch.stack([P0, P1], dim=0)  # (2 N)
        P *= 2 ** (
            -0.5
        )  # Halve the rank correct just like the original matrix was halved
    elif measure in ["fourier", "fout"]:
        P = torch.zeros(N)
        P[0::2] = 2**0.5
        P[0] = 1
        P = P.unsqueeze(0)
    elif measure in ["fourier_diag", "foud", "legsd"]:
        P = torch.zeros(1, N, dtype=dtype)
    else:
        raise NotImplementedError

    d = P.size(0)
    if rank > d:
        P = torch.cat([P, torch.zeros(rank - d, N, dtype=dtype)], dim=0)  # (rank N)
    return P


def nplr(measure, N, rank=1, dtype=torch.float, diagonalize_precision=True):
    """Return w, p, q, V, B such that
    (w - p q^*, B) is unitarily equivalent to the original HiPPO A, B by the matrix V
    i.e. A = V[w - p q^*]V^*, B = V B
    """
    assert dtype == torch.float or torch.double
    cdtype = torch.cfloat if dtype == torch.float else torch.cdouble

    A, B = transition(measure, N)
    A = torch.as_tensor(A, dtype=dtype)  # (N, N)
    B = torch.as_tensor(B, dtype=dtype)[:, 0]  # (N,)

    P = rank_correction(measure, N, rank=rank, dtype=dtype)  # (r N)
    AP = A + torch.sum(P.unsqueeze(-2) * P.unsqueeze(-1), dim=-3)

    # We require AP to be nearly skew-symmetric
    _A = AP + AP.transpose(-1, -2)
    if (
        err := torch.sum((_A - _A[0, 0] * torch.eye(N)) ** 2) / N
    ) > 1e-5:  # if not torch.allclose(_A - _A[0,0]*torch.eye(N), torch.zeros(N, N), atol=1e-5):
        print("WARNING: HiPPO matrix not skew symmetric", err)

    # Take advantage of identity + skew-symmetric form to calculate real and imaginary parts separately
    # Imaginary part can use eigh instead of eig
    w_re = torch.mean(torch.diagonal(AP), -1, keepdim=True)

    # Diagonalize in double precision
    if diagonalize_precision:
        AP = AP.to(torch.double)
    w_im, V = torch.linalg.eigh(AP * -1j)  # (..., N) (..., N, N)
    if diagonalize_precision:
        w_im, V = w_im.to(cdtype), V.to(cdtype)
    w = w_re + 1j * w_im
    # Check: V w V^{-1} = A
    # print("check", V @ torch.diag_embed(w) @ V.conj().transpose(-1, -2))

    # Only keep half of each conjugate pair
    _, idx = torch.sort(w.imag)
    w_sorted = w[idx]
    V_sorted = V[:, idx]

    # There is an edge case when eigenvalues can be 0, which requires some machinery to handle
    # We use a huge hack here: Assume only one pair is 0, and that it is the first row/column of A (only happens in Fourier case)
    V = V_sorted[:, : N // 2]
    w = w_sorted[: N // 2]
    assert w[-2].abs() > 1e-4, "Only 1 zero eigenvalue allowed in diagonal part of A"
    if w[-1].abs() < 1e-4:
        V[:, -1] = 0.0
        V[0, -1] = 2**-0.5
        V[1, -1] = 2**-0.5 * 1j

    _AP = V @ torch.diag_embed(w) @ V.conj().transpose(-1, -2)
    if (err := torch.sum((2 * _AP.real - AP) ** 2) / N) > 1e-5:
        print(
            "Warning: Diagonalization of A matrix not numerically precise - error", err
        )
    # print("check", V @ torch.diag_embed(w) @ V.conj().transpose(-1, -2))

    V_inv = V.conj().transpose(-1, -2)

    B = contract("ij, j -> i", V_inv, B.to(V))  # V^* B
    P = contract("ij, ...j -> ...i", V_inv, P.to(V))  # V^* P

    return w, P, B, V


def dplr(
    scaling,
    N,
    rank=1,
    H=1,
    dtype=torch.float,
    real_scale=1.0,
    imag_scale=1.0,
    random_real=False,
    random_imag=False,
    normalize=False,
    diagonal=True,
    random_B=False,
):
    assert dtype == torch.float or torch.double
    dtype = torch.cfloat if dtype == torch.float else torch.cdouble

    pi = torch.tensor(math.pi)
    if random_real:
        real_part = torch.rand(H, N // 2)
    else:
        real_part = 0.5 * torch.ones(H, N // 2)
    if random_imag:
        imag_part = N // 2 * torch.rand(H, N // 2)
    else:
        imag_part = repeat(torch.arange(N // 2), "n -> h n", h=H)

    real_part = real_scale * real_part
    if scaling == "random":
        imag_part = torch.randn(H, N // 2)
    elif scaling == "real":
        imag_part = 0 * imag_part
        real_part = 1 + repeat(torch.arange(N // 2), "n -> h n", h=H)
    elif scaling in ["linear", "lin"]:
        imag_part = pi * imag_part
    elif scaling in [
        "inverse",
        "inv",
    ]:  # Based on asymptotics of the default HiPPO matrix
        imag_part = 1 / pi * N * (N / (1 + 2 * imag_part) - 1)
    elif scaling in ["inverse2", "inv2"]:
        imag_part = 1 / pi * N * (N / (1 + imag_part) - 1)
    elif scaling in ["quadratic", "quad"]:
        imag_part = 1 / pi * (1 + 2 * imag_part) ** 2
    elif scaling in ["legs", "hippo"]:
        w, _, _, _ = nplr("legsd", N)
        imag_part = w.imag

    else:
        raise NotImplementedError
    imag_part = imag_scale * imag_part
    w = -real_part + 1j * imag_part

    # Initialize B
    if random_B:
        B = torch.randn(H, N // 2, dtype=dtype)
    else:
        B = torch.ones(H, N // 2, dtype=dtype)

    if normalize:
        norm = (
            -B / w
        )  # (H, N) # Result if you integrate the kernel with constant 1 function
        zeta = 2 * torch.sum(
            torch.abs(norm) ** 2, dim=-1, keepdim=True
        )  # Variance with a random C vector
        B = B / zeta**0.5

    P = torch.randn(rank, H, N // 2, dtype=dtype)
    if diagonal:
        P = P * 0.0
    V = torch.eye(N, dtype=dtype)[:: N // 2]  # Only used in testing
    V = repeat(V, "n m -> h n m", h=H)

    return w, P, B, V


def ssm(measure, N, R, H, **ssm_args):
    """Dispatcher to create single SSM initialization

    N: state size
    R: rank (for DPLR parameterization)
    H: number of independent SSM copies
    """

    if measure == "dplr":
        w, P, B, V = dplr(N=N, rank=R, H=H, **ssm_args)
    elif measure.startswith("diag"):
        args = measure.split("-")
        assert args[0] == "diag" and len(args) > 1
        scaling = args[1]
        w, P, B, V = dplr(scaling=scaling, N=N, rank=R, H=H, diagonal=True, **ssm_args)
    else:
        w, P, B, V = nplr(measure, N, R, **ssm_args)
        w = repeat(w, "n -> s n", s=H)
        P = repeat(P, "r n -> r s n", s=H)
        B = repeat(B, "n -> s n", s=H)
        V = repeat(V, "n m -> s n m", s=H)
    return w, P, B, V


combinations = {
    "hippo": ["legs", "fourier"],
    "diag": ["diag-inv", "diag-lin"],
    "all": ["legs", "fourier", "diag-inv", "diag-lin"],
}


def combination(measures, N, R, S, **ssm_args):
    if isinstance(measures, str):
        measures = combinations[measures] if measures in combinations else [measures]

    assert S % len(measures) == 0, (
        f"{S} independent trainable SSM copies must be multiple of"
        f" {len(measures)} different measures"
    )
    w, P, B, V = zip(
        *[ssm(measure, N, R, S // len(measures), **ssm_args) for measure in measures]
    )
    w = torch.cat(w, dim=0)  # (S N)
    P = torch.cat(P, dim=1)  # (R S N)
    B = torch.cat(B, dim=0)  # (S N)
    V = torch.cat(V, dim=0)  # (S N N)
    return w, P, B, V


class OptimModule(nn.Module):
    """Interface for Module that allows registering buffers/parameters with configurable optimizer hyperparameters"""

    def register(self, name, tensor, lr=None):
        """Register a tensor with a configurable learning rate and 0 weight decay"""

        if lr == 0.0:
            self.register_buffer(name, tensor)
        else:
            self.register_parameter(name, nn.Parameter(tensor))

            optim = {"weight_decay": 0.0}
            if lr is not None:
                optim["lr"] = lr
            setattr(getattr(self, name), "_optim", optim)


class SSKernelNPLR(OptimModule):
    """Stores a representation of and computes the SSKernel function K_L(A^dt, B^dt, C) corresponding to a discretized state space, where A is Normal + Low Rank (NPLR)"""

    @torch.no_grad()
    def _setup_C(self, L):
        """Construct C~ from C

        Two modes are supported: go directly to length L if self.L is 1, or length is doubled
        """

        if self.L.item() == 0:
            if self.verbose:
                log.info(f"S4: Initializing kernel to length {L}")
            double_length = False
        elif L > self.L.item():  # 2*int(self.L) == L:
            if self.verbose:
                log.info(
                    f"S4: Doubling length from L = {self.L.item()} to {2*self.L.item()}"
                )
            double_length = True
            L = self.L.item()  # Convenience for the math below
        else:
            return

        C = _r2c(self.C)
        dA, _ = self._setup_state()
        dA_L = power(L, dA)
        # Multiply C by I - dA_L
        C_ = _conj(C)
        prod = contract("h m n, c h n -> c h m", dA_L.transpose(-1, -2), C_)
        if double_length:
            prod = -prod  # Multiply by I + dA_L instead
        C_ = C_ - prod
        C_ = C_[..., : self.N]  # Take conjugate pairs again
        self.C.copy_(_c2r(C_))

        self.L = 2 * self.L if double_length else self.L + L  # Preserve type/device

    def _omega(self, L, dtype, device, cache=True):
        """Calculate (and cache) FFT nodes and their "unprocessed" version with the bilinear transform
        This should be called everytime the internal length self.L changes"""

        # Use cached if available
        if cache and hasattr(self, "omega") and self.omega.size(-1) == L // 2 + 1:
            return self.omega, self.z

        omega = torch.tensor(
            np.exp(-2j * np.pi / (L)), dtype=dtype, device=device
        )  # \omega_{2L}
        omega = omega ** torch.arange(0, L // 2 + 1, device=device)
        z = 2 * (1 - omega) / (1 + omega)

        # Cache if necessary
        if cache:
            self.omega = omega
            self.z = z
        return omega, z

    def __init__(
        self,
        w,
        P,
        B,
        C,
        log_dt,
        L=None,  # starting/maximum length of kernel
        lr=None,
        verbose=False,
        keops=False,
        real_type="exp",  # ['none' | 'exp' | 'relu' | sigmoid']
        real_tolerance=1e-3,
        bandlimit=None,
    ):
        """
        L: Maximum length; this module computes an SSM kernel of length L
        A is represented by diag(w) - PP^*
        w: (S, N) diagonal part
        P: (R, S, N) low-rank part

        B: (S, N)
        C: (C, H, N)
        dt: (H) timescale per feature
        lr: [dict | float | None] hook to set lr of special parameters (A, B, dt)

        Dimensions:
        N (or d_state): state size
        H (or d_model): total SSM copies
        S (or n_ssm): number of trainable copies of (A, B, dt); must divide H
        R (or rank): rank of low-rank part
        C (or channels): system is 1-dim to C-dim

        The forward pass of this Module returns a tensor of shape (C, H, L)

        Note: tensor shape N here denotes half the true state size, because of conjugate symmetry
        """

        super().__init__()
        self.verbose = verbose
        self.keops = keops
        self.bandlimit = bandlimit
        self.real_type = real_type
        self.real_tolerance = real_tolerance

        # Rank of low-rank correction
        self.rank = P.shape[-3]
        assert w.size(-1) == P.size(-1) == B.size(-1) == C.size(-1)
        self.H = log_dt.size(-1)
        self.N = w.size(-1)

        # Check different SSM inits
        assert w.size(-2) == P.size(-2) == B.size(-2)  # n_ssm
        assert self.H % w.size(0) == 0
        self.n_ssm = w.size(0)
        self.broadcast = self.H // w.size(
            0
        )  # Each trainable SSM needs to be duplicated this many times

        # Broadcast everything to correct shapes
        C = C.expand(torch.broadcast_shapes(C.shape, (1, self.H, self.N)))  # (C, H, N)
        B = B.unsqueeze(0)  # (1, 1, N)

        # Register parameters
        self.C = nn.Parameter(_c2r(_resolve_conj(C)))
        if lr is None or isinstance(lr, float):
            lr_dict = {}
        else:
            lr_dict, lr = lr, None
        self.register("log_dt", log_dt, lr_dict.get("dt", lr))
        self.register("B", _c2r(B), lr_dict.get("B", lr))
        self.register("P", _c2r(P), lr_dict.get("A", lr))
        self.register("inv_w_real", self._w_init(w.real), lr_dict.get("A", lr))
        self.register("w_imag", w.imag, lr_dict.get("A", lr))

        self.l_max = L
        self.register_buffer("L", torch.tensor(0))  # Internal length

    def _w_init(self, w_real):
        w_real = torch.clamp(w_real, max=-self.real_tolerance)
        if self.real_type == "none":
            return -w_real
        elif self.real_type == "exp":
            return torch.log(-w_real)  # Some of the HiPPO methods have real part 0
        elif self.real_type == "relu":
            return -w_real
        elif self.real_type == "sigmoid":
            return torch.logit(-w_real)
        elif self.real_type == "softplus":
            return torch.log(torch.exp(-w_real) - 1)
        else:
            raise NotImplementedError

    def _w(self):
        # Get the internal w (diagonal) parameter
        if self.real_type == "none":
            w_real = -self.inv_w_real
        elif self.real_type == "exp":
            w_real = -torch.exp(self.inv_w_real)
        elif self.real_type == "relu":
            w_real = -F.relu(self.inv_w_real)
        elif self.real_type == "sigmoid":
            w_real = -F.sigmoid(self.inv_w_real)
        elif self.real_type == "softplus":
            w_real = -F.softplus(self.inv_w_real)
        else:
            raise NotImplementedError
        w = w_real + 1j * self.w_imag
        return w

    def forward(self, state=None, rate=1.0, L=None):
        """
        state: (B, H, N) initial state
        rate: sampling rate factor
        L: target length

        returns:
        (C, H, L) convolution kernel (generally C=1)
        (B, H, L) output from initial state
        """

        # Initialize C~ if necessary (done in forward pass so it's on the correct device)
        if self.L.item() == 0 and self.l_max is not None and self.l_max > 0:
            self._setup_C(self.l_max)

        # Handle sampling rate logic
        # The idea is that this kernel's length (in continuous units) is self.L, while we are asked to provide a kernel of length L at (relative) frequency rate
        if L is None:
            L = round(self.L.item() / rate)

        # Increase the internal length if needed
        continuous_L = round(rate * L)
        while continuous_L > self.L.item():
            self._setup_C(continuous_L)
        discrete_L = round(self.L.item() / rate)

        dt = torch.exp(self.log_dt) * rate
        B = _r2c(self.B)
        C = _r2c(self.C)
        P = _r2c(self.P)
        Q = P.conj()
        w = self._w()  # (n_ssm, N)

        # Address bandlimiting
        if self.bandlimit is not None:
            freqs = w.imag.abs() / (2 * math.pi)  # (H, N)
            freqs = dt[:, None] / rate * freqs  # (H, N)
            mask = torch.where(freqs < self.bandlimit * 0.5, 1, 0)
            C = C * mask

        # Get FFT nodes of right length
        omega, z = self._omega(
            discrete_L, dtype=w.dtype, device=w.device, cache=(rate == 1.0)
        )

        # Broadcast parameters to same hidden features H
        B = repeat(B, "1 t n -> 1 (v t) n", v=self.broadcast)
        P = repeat(P, "r t n -> r (v t) n", v=self.broadcast)
        Q = repeat(Q, "r t n -> r (v t) n", v=self.broadcast)
        w = repeat(w, "t n -> (v t) n", v=self.broadcast)

        # Augment B
        if state is not None:
            # Have to "unbilinear" the state to put it into the same "type" as B
            # Compute 1/dt * (I + dt/2 A) @ state

            # Can do this without expanding (maybe minor speedup using conj symmetry in theory), but it's easier to read this way
            s = _conj(state) if state.size(-1) == self.N else state  # (B H N)
            sA = s * _conj(w) - contract(  # (B H N)
                "bhm, rhm, rhn -> bhn", s, _conj(Q), _conj(P)
            )
            s = s / dt.unsqueeze(-1) + sA / 2
            s = s[..., : self.N]

            B = torch.cat([s, B], dim=-3)  # (B+1, H, N)

        # Incorporate dt into A
        w = w * dt.unsqueeze(-1)  # (H N)

        # Stack B and p, C and q for convenient batching
        B = torch.cat([B, P], dim=-3)  # (B+1+R, H, N)
        C = torch.cat([C, Q], dim=-3)  # (C+R, H, N)

        # Incorporate B and C batch dimensions
        v = B.unsqueeze(-3) * C.unsqueeze(-4)  # (B+1+R, C+R, H, N)

        # Calculate resolvent at omega
        if has_cauchy_extension and z.dtype == torch.cfloat and not self.keops:
            r = cauchy_mult(v, z, w, symmetric=True)
        elif has_pykeops:
            r = cauchy_conj(v, z, w)
        else:
            r = cauchy_naive(v, z, w)
        r = r * dt[None, None, :, None]  # (B+1+R, C+R, H, L)

        # Low-rank Woodbury correction
        if self.rank == 1:
            k_f = r[:-1, :-1, :, :] - r[:-1, -1:, :, :] * r[-1:, :-1, :, :] / (
                1 + r[-1:, -1:, :, :]
            )
        elif self.rank == 2:
            r00 = r[: -self.rank, : -self.rank, :, :]
            r01 = r[: -self.rank, -self.rank :, :, :]
            r10 = r[-self.rank :, : -self.rank, :, :]
            r11 = r[-self.rank :, -self.rank :, :, :]
            det = (1 + r11[:1, :1, :, :]) * (1 + r11[1:, 1:, :, :]) - r11[
                :1, 1:, :, :
            ] * r11[1:, :1, :, :]
            s = (
                r01[:, :1, :, :] * (1 + r11[1:, 1:, :, :]) * r10[:1, :, :, :]
                + r01[:, 1:, :, :] * (1 + r11[:1, :1, :, :]) * r10[1:, :, :, :]
                - r01[:, :1, :, :] * (r11[:1, 1:, :, :]) * r10[1:, :, :, :]
                - r01[:, 1:, :, :] * (r11[1:, :1, :, :]) * r10[:1, :, :, :]
            )
            s = s / det
            k_f = r00 - s
        else:
            r00 = r[: -self.rank, : -self.rank, :, :]
            r01 = r[: -self.rank, -self.rank :, :, :]
            r10 = r[-self.rank :, : -self.rank, :, :]
            r11 = r[-self.rank :, -self.rank :, :, :]
            r11 = rearrange(r11, "a b h n -> h n a b")
            r11 = torch.linalg.inv(torch.eye(self.rank, device=r.device) + r11)
            r11 = rearrange(r11, "h n a b -> a b h n")
            k_f = r00 - torch.einsum(
                "i j h n, j k h n, k l h n -> i l h n", r01, r11, r10
            )

        # Final correction for the bilinear transform
        k_f = k_f * 2 / (1 + omega)

        # Move from frequency to coefficients
        k = torch.fft.irfft(k_f, n=discrete_L)  # (B+1, C, H, L)

        # # Truncate to target length
        k = k[..., :L]

        if state is not None:
            k_state = k[:-1, :, :, :]  # (B, C, H, L)
        else:
            k_state = None
        k_B = k[-1, :, :, :]  # (C H L)

        return k_B, k_state

    @torch.no_grad()
    def _setup_linear(self):
        """Create parameters that allow fast linear stepping of state"""
        w = self._w()
        B = _r2c(self.B)  # (H N)
        P = _r2c(self.P)
        Q = P.conj()

        # Repeat w shape properly
        B = repeat(B, "1 t n -> 1 (v t) n", v=self.broadcast)
        P = repeat(P, "r t n -> r (v t) n", v=self.broadcast)
        Q = repeat(Q, "r t n -> r (v t) n", v=self.broadcast)
        w = repeat(w, "t n -> (v t) n", v=self.broadcast)

        # Prepare Linear stepping
        dt = torch.exp(self.log_dt)
        D = (2.0 / dt.unsqueeze(-1) - w).reciprocal()  # (H, N)
        R = (
            torch.eye(self.rank, dtype=w.dtype, device=w.device)
            + 2 * contract("r h n, h n, s h n -> h r s", Q, D, P).real
        )  # (H R R)
        Q_D = rearrange(Q * D, "r h n -> h r n")
        try:
            R = torch.linalg.solve(R, Q_D)  # (H R N)
        except:
            R = torch.tensor(
                np.linalg.solve(
                    R.to(Q_D).contiguous().detach().cpu(),
                    Q_D.contiguous().detach().cpu(),
                )
            ).to(Q_D)
        R = rearrange(R, "h r n -> r h n")

        self.step_params = {
            "D": D,  # (H N)
            "R": R,  # (R H N)
            "P": P,  # (R H N)
            "Q": Q,  # (R H N)
            "B": B,  # (1 H N)
            "E": 2.0 / dt.unsqueeze(-1) + w,  # (H N)
        }

    def _step_state_linear(self, u=None, state=None):
        """
        Version of the step function that has time O(N) instead of O(N^2) per step, which takes advantage of the DPLR form and bilinear discretization.

        Unfortunately, as currently implemented it's about 2x slower because it calls several sequential operations. Perhaps a fused CUDA kernel implementation would be much faster

        u: (H) input
        state: (H, N/2) state with conjugate pairs
          Optionally, the state can have last dimension N
        Returns: same shape as state
        """
        C = _r2c(self.C)  # View used for dtype/device

        if u is None:  # Special case used to find dA
            u = torch.zeros(self.H, dtype=C.dtype, device=C.device)
        if state is None:  # Special case used to find dB
            state = torch.zeros(self.H, self.N, dtype=C.dtype, device=C.device)

        step_params = self.step_params.copy()
        if (
            state.size(-1) == self.N
        ):  # Only store half of the conjugate pairs; should be true by default
            # There should be a slightly faster way using conjugate symmetry
            contract_fn = lambda p, x, y: contract(
                "r h n, r h m, ... h m -> ... h n", _conj(p), _conj(x), _conj(y)
            )[
                ..., : self.N
            ]  # inner outer product
        else:
            assert state.size(-1) == 2 * self.N
            step_params = {k: _conj(v) for k, v in step_params.items()}
            # TODO worth setting up a contract_expression in default_state if we want to use this at inference time for stepping
            contract_fn = lambda p, x, y: contract(
                "r h n, r h m, ... h m -> ... h n", p, x, y
            )  # inner outer product
        D = step_params["D"]  # (H N)
        E = step_params["E"]  # (H N)
        R = step_params["R"]  # (R H N)
        P = step_params["P"]  # (R H N)
        Q = step_params["Q"]  # (R H N)
        B = step_params["B"]  # (1 H N)

        new_state = E * state - contract_fn(P, Q, state)  # (B H N)
        new_state = new_state + 2.0 * B * u.unsqueeze(-1)  # (B H N)
        new_state = D * (new_state - contract_fn(P, R, new_state))

        return new_state

    def _setup_state(self):
        """Construct dA and dB for discretized state equation"""

        # Construct dA and dB by using the stepping
        self._setup_linear()
        C = _r2c(self.C)  # Just returns a view that we use for finding dtype/device

        state = torch.eye(2 * self.N, dtype=C.dtype, device=C.device).unsqueeze(
            -2
        )  # (N 1 N)
        dA = self._step_state_linear(state=state)
        dA = rearrange(dA, "n h m -> h m n")

        u = C.new_ones(self.H)
        dB = self._step_state_linear(u=u)
        dB = _conj(dB)
        dB = rearrange(dB, "1 h n -> h n")  # (H N)
        return dA, dB

    def _step_state(self, u, state):
        """Must be called after self.default_state() is used to construct an initial state!"""
        next_state = self.state_contraction(self.dA, state) + self.input_contraction(
            self.dB, u
        )
        return next_state

    def _setup_step(self, mode="dense"):
        """Set up dA, dB, dC discretized parameters for stepping"""
        self.dA, self.dB = self._setup_state()

        # Calculate original C
        C = _conj(_r2c(self.C))  # (H C N)
        if self.L.item() == 0:
            dC = C
        else:
            # self.C represents C_tilde
            dA_L = power(self.L.item(), self.dA)
            I = torch.eye(self.dA.size(-1)).to(dA_L)

            dC = torch.linalg.solve(
                I - dA_L.transpose(-1, -2),
                C.unsqueeze(-1),
            ).squeeze(-1)
        self.dC = dC

        # Do special preprocessing for different step modes

        self._step_mode = mode
        if mode == "linear":
            # Linear case: special step function for the state, we need to handle output
            # use conjugate symmetry by default, which affects the output projection
            self.dC = 2 * self.dC[:, :, : self.N]
        elif mode == "diagonal":
            # Eigendecomposition of the A matrix
            L, V = torch.linalg.eig(self.dA)
            V_inv = torch.linalg.inv(V)
            # Check that the eigendedecomposition is correct
            if self.verbose:
                print(
                    "Diagonalization error:",
                    torch.dist(V @ torch.diag_embed(L) @ V_inv, self.dA),
                )

            # Change the parameterization to diagonalize
            self.dA = L
            self.dB = contract("h n m, h m -> h n", V_inv, self.dB)
            self.dC = contract("h n m, c h n -> c h m", V, self.dC)

        elif mode == "dense":
            pass
        else:
            raise NotImplementedError(
                "NPLR Kernel step mode must be {'dense' | 'linear' | 'diagonal'}"
            )

    def default_state(self, *batch_shape):
        C = _r2c(self.C)
        N = C.size(-1)
        H = C.size(-2)

        # Cache the tensor contractions we will later do, for efficiency
        # These are put in this function because they depend on the batch size
        step_mode = getattr(
            self, "_step_mode", "dense"
        )  # Used in default_state, which is called without _setup_step() in forward_state()
        if step_mode != "linear":
            N *= 2

            if step_mode == "diagonal":
                self.state_contraction = contract_expression(
                    "h n, ... h n -> ... h n",
                    (H, N),
                    batch_shape + (H, N),
                )
            else:
                # Dense (quadratic) case: expand all terms
                self.state_contraction = contract_expression(
                    "h m n, ... h n -> ... h m",
                    (H, N, N),
                    batch_shape + (H, N),
                )

            self.input_contraction = contract_expression(
                "h n, ... h -> ... h n",
                (H, N),  # self.dB.shape
                batch_shape + (H,),
            )

        self.output_contraction = contract_expression(
            "c h n, ... h n -> ... c h",
            (C.shape[0], H, N),  # self.dC.shape
            batch_shape + (H, N),
        )

        state = torch.zeros(*batch_shape, H, N, dtype=C.dtype, device=C.device)
        return state

    def step(self, u, state):
        """Must have called self._setup_step() and created state with self.default_state() before calling this"""

        if self._step_mode == "linear":
            new_state = self._step_state_linear(u, state)
        else:
            new_state = self._step_state(u, state)
        y = self.output_contraction(self.dC, new_state)
        return y.real, new_state


class SSKernelDiag(OptimModule):
    """Version using (complex) diagonal state matrix (S4D)"""

    def __init__(
        self,
        A,
        B,
        C,
        log_dt,
        L=None,
        disc="bilinear",
        real_type="exp",
        lr=None,
        bandlimit=None,
    ):

        super().__init__()
        self.L = L
        self.disc = disc
        self.bandlimit = bandlimit
        self.real_type = real_type

        # Rank of low-rank correction
        assert A.size(-1) == C.size(-1)
        self.H = log_dt.size(-1)
        self.N = A.size(-1)
        assert A.size(-2) == B.size(-2)  # Number of independent SSMs trained
        assert self.H % A.size(-2) == 0
        self.n_ssm = A.size(-2)
        self.repeat = self.H // A.size(0)

        self.channels = C.shape[0]
        self.C = nn.Parameter(_c2r(_resolve_conj(C)))

        # Register parameters
        if lr is None or isinstance(lr, float):
            lr_dict = {}
        else:
            lr_dict, lr = lr, None

        self.register("log_dt", log_dt, lr_dict.get("dt", lr))
        self.register("A", _c2r(A), lr_dict.get("A", lr))
        self.register("B", _c2r(B), lr_dict.get("B", lr))
        self.register("inv_A_real", self._A_init(A.real), lr_dict.get("A", lr))
        self.register("A_imag", A.imag, lr_dict.get("A", lr))

    def _A_init(self, A_real):
        A_real = torch.clamp(A_real, max=-1e-4)
        if self.real_type == "none":
            return -A_real
        elif self.real_type == "exp":
            return torch.log(-A_real)  # Some of the HiPPO methods have real part 0
        elif self.real_type == "relu":
            return -A_real
        elif self.real_type == "sigmoid":
            return torch.logit(-A_real)
        elif self.real_type == "softplus":
            return torch.log(torch.exp(-A_real) - 1)
        else:
            raise NotImplementedError

    def _A(self):
        # Get the internal A (diagonal) parameter
        if self.real_type == "none":
            A_real = -self.inv_A_real
        elif self.real_type == "exp":
            A_real = -torch.exp(self.inv_A_real)
        elif self.real_type == "relu":
            # JAX version seems to NaN if you alloA 0's, although this code Aas fine Aithout it
            A_real = -F.relu(self.inv_A_real) - 1e-4
        elif self.real_type == "sigmoid":
            A_real = -F.sigmoid(self.inv_A_real)
        elif self.real_type == "softplus":
            A_real = -F.softplus(self.inv_A_real)
        else:
            raise NotImplementedError
        A = A_real + 1j * self.A_imag
        return A

    def forward(self, L, state=None, rate=1.0, u=None):
        """
        state: (B, H, N) initial state
        rate: sampling rate factor
        L: target length

        returns:
        (C, H, L) convolution kernel (generally C=1)
        (B, H, L) output from initial state
        """

        dt = torch.exp(self.log_dt) * rate  # (H)
        C = _r2c(self.C)  # (C H N)
        A = self._A()  # (H N)

        B = _r2c(self.B)
        B = repeat(B, "t n -> 1 (v t) n", v=self.repeat)

        if self.bandlimit is not None:
            freqs = dt[:, None] / rate * A.imag.abs() / (2 * math.pi)  # (H, N)
            mask = torch.where(freqs < self.bandlimit * 0.5, 1, 0)
            C = C * mask

        # Incorporate dt into A
        A = repeat(A, "t n -> (v t) n", v=self.repeat)
        dtA = A * dt.unsqueeze(-1)  # (H N)

        # Augment B with state
        if state is not None:
            s = state / dt.unsqueeze(-1)
            if self.disc == "bilinear":
                s = s * (1.0 + dtA / 2)
            elif self.disc == "zoh":
                s = s * dtA * dtA.exp() / (dtA.exp() - 1.0)
            B = torch.cat([s, B], dim=-3)  # (1+B H N)

        C = (B[:, None, :, :] * C).view(-1, self.H, self.N)
        if self.disc == "zoh":
            # Power up
            C = C * (torch.exp(dtA) - 1.0) / A
            K = log_vandermonde(C, dtA, L)  # (H L)
        elif self.disc == "bilinear":
            C = C * (1.0 - dtA / 2).reciprocal() * dt.unsqueeze(-1)  # or * dtA / A
            dA = (1.0 + dtA / 2) / (1.0 - dtA / 2)
            K = log_vandermonde(C, dA.log(), L)
        elif self.disc == "dss":
            # Implementation from DSS meant for case when real eigenvalues can be positive
            P = dtA.unsqueeze(-1) * torch.arange(L, device=C.device)  # [H N L]
            A_gt_0 = A.real > 0  # [N]
            if A_gt_0.any():
                with torch.no_grad():
                    P_max = dtA * (A_gt_0 * (L - 1))  # [H N]
                P = P - P_max.unsqueeze(-1)  # [H N L]
            S = P.exp()  # [H N L]

            dtA_neg = dtA * (1 - 2 * A_gt_0)  # [H N]
            num = dtA_neg.exp() - 1  # [H N]
            den = (dtA_neg * L).exp() - 1  # [H N]

            # Inline reciprocal function for DSS logic
            x = den * A
            x_conj = _resolve_conj(x)
            r = x_conj / (x * x_conj + 1e-7)

            C = C * num * r  # [C H N]
            K = contract("chn,hnl->chl", C, S).float()
        else:
            assert False, f"{self.disc} not supported"

        K = K.view(-1, self.channels, self.H, L)  # (1+B C H L)
        if state is not None:
            K_state = K[:-1, :, :, :]  # (B C H L)
        else:
            K_state = None
        K = K[-1, :, :, :]  # (C H L)
        return K, K_state

    def _setup_step(self):
        # These methods are organized like this to be compatible with the NPLR kernel interface
        dt = torch.exp(self.log_dt)  # (H)
        B = _r2c(self.B)  # (H N)
        C = _r2c(self.C)  # (C H N)
        self.dC = C
        A = self._A()  # (H N)

        # Incorporate dt into A
        dtA = A * dt.unsqueeze(-1)  # (H N)
        if self.disc == "zoh":
            self.dA = torch.exp(dtA)  # (H N)
            self.dB = B * (torch.exp(dtA) - 1.0) / A  # (C H N)
        elif self.disc == "bilinear":
            self.dA = (1.0 + dtA / 2) / (1.0 - dtA / 2)
            self.dB = (
                B * (1.0 - dtA / 2).reciprocal() * dt.unsqueeze(-1)
            )  # or * dtA / A

    def default_state(self, *batch_shape):
        C = _r2c(self.C)
        state = torch.zeros(
            *batch_shape, self.H, self.N, dtype=C.dtype, device=C.device
        )
        return state

    def step(self, u, state):
        next_state = contract("h n, b h n -> b h n", self.dA, state) + contract(
            "h n, b h -> b h n", self.dB, u
        )
        y = contract("c h n, b h n -> b c h", self.dC, next_state)
        return 2 * y.real, next_state

    def forward_state(self, u, state):
        self._setup_step()
        AL = self.dA ** u.size(-1)
        u = u.flip(-1).to(self.dA).contiguous()  # (B H L)
        v = log_vandermonde_transpose(u, self.dB, self.dA.log(), u.size(-1))
        next_state = AL * state + v
        return next_state


class SSKernel(nn.Module):
    """Wrapper around SSKernel parameterizations.

    The SSKernel is expected to support the interface
    forward()
    default_state()
    _setup_step()
    step()
    """

    def __init__(
        self,
        H,
        N=64,
        L=None,
        measure="legs",
        rank=1,
        channels=1,
        dt_min=0.001,
        dt_max=0.1,
        deterministic=False,
        lr=None,
        mode="nplr",
        n_ssm=None,
        verbose=False,
        measure_args={},
        **kernel_args,
    ):
        """State Space Kernel which computes the convolution kernel $\\bar{K}$

        H: Number of independent SSM copies; controls the size of the model. Also called d_model in the config.
        N: State size (dimensionality of parameters A, B, C). Also called d_state in the config. Generally shouldn't need to be adjusted and doens't affect speed much.
        L: Maximum length of convolution kernel, if known. Should work in the majority of cases even if not known.
        measure: Options for initialization of (A, B). For NPLR mode, recommendations are "legs", "fout", "hippo" (combination of both). For Diag mode, recommendations are "diag-inv", "diag-lin", "diag-legs", and "diag" (combination of diag-inv and diag-lin)
        rank: Rank of low-rank correction for NPLR mode. Needs to be increased for measure "legt"
        channels: C channels turns the SSM from a 1-dim to C-dim map; can think of it having C separate "heads" per SSM. This was partly a feature to make it easier to implement bidirectionality; it is recommended to set channels=1 and adjust H to control parameters instead
        dt_min, dt_max: min and max values for the step size dt (\Delta)
        mode: Which kernel algorithm to use. 'nplr' is the full S4 model; 'diag' is the simpler S4D; 'slow' is a dense version for testing
        n_ssm: Number of independent trainable (A, B) SSMs, e.g. n_ssm=1 means all A/B parameters are tied across the H different instantiations of C. n_ssm=None means all H SSMs are completely independent. Generally, changing this option can save parameters but doesn't affect performance or speed much. This parameter must divide H
        lr: Passing in a number (e.g. 0.001) sets attributes of SSM parameers (A, B, dt). A custom optimizer hook is needed to configure the optimizer to set the learning rates appropriately for these parameters.
        """
        super().__init__()
        self.N = N
        self.H = H
        dtype, cdtype = torch.float, torch.cfloat
        self.channels = channels
        self.n_ssm = n_ssm if n_ssm is not None else H
        self.mode = mode
        self.verbose = verbose
        self.kernel_args = kernel_args

        # Generate dt
        if deterministic:
            log_dt = torch.exp(torch.linspace(math.log(dt_min), math.log(dt_max), H))
        else:
            log_dt = torch.rand(self.H, dtype=dtype) * (
                math.log(dt_max) - math.log(dt_min)
            ) + math.log(dt_min)

        # Compute the preprocessed representation
        w, P, B, V = combination(measure, self.N, rank, self.n_ssm, **measure_args)

        # Broadcast C to have H channels
        if deterministic:
            C = torch.zeros(channels, self.H, self.N, dtype=cdtype)
            C[:, :, :1] = 1.0
            C = contract("hmn, chn -> chm", V.conj().transpose(-1, -2), C)  # V^* C
        else:
            C = torch.randn(channels, self.H, self.N // 2, dtype=cdtype)

        # Broadcast other parameters to have n_ssm copies
        assert (
            self.n_ssm % B.size(-2) == 0
            and self.n_ssm % P.size(-2) == 0
            and self.n_ssm % w.size(-2) == 0
        )
        # Broadcast tensors to n_ssm copies
        # These will be the parameters, so make sure tensors are materialized and contiguous
        B = repeat(B, "t n -> (v t) n", v=self.n_ssm // B.size(-2)).clone().contiguous()
        P = (
            repeat(P, "r t n -> r (v t) n", v=self.n_ssm // P.size(-2))
            .clone()
            .contiguous()
        )
        w = repeat(w, "t n -> (v t) n", v=self.n_ssm // w.size(-2)).clone().contiguous()
        C = C.contiguous()

        if mode == "nplr":
            self.kernel = SSKernelNPLR(
                w,
                P,
                B,
                C,
                log_dt,
                L=L,
                lr=lr,
                verbose=verbose,
                **kernel_args,
            )
        elif mode == "diag":
            if not measure.startswith("diag"):
                log.warning(
                    "Diagonal kernel (S4D) activated but initialization is not intended"
                    " for S4D. Set `measure` to 'diag-lin', 'diag-inv', or 'diag-legs'"
                    " for the main variants, or 'diag' for a combination of S4D-Lin and"
                    " S4D-Inv."
                )
            C = C * repeat(B, "t n -> (v t) n", v=H // self.n_ssm)
            self.kernel = SSKernelDiag(
                w,
                B,
                C,
                log_dt,
                L=L,
                lr=lr,
                **kernel_args,
            )
        else:
            raise NotImplementedError(f"{mode=} is not valid")

    def forward(self, state=None, L=None, rate=None):
        return self.kernel(state=state, L=L, rate=rate)

    @torch.no_grad()
    def forward_state(self, u, state):
        """Forward the state through a sequence, i.e. computes the state after passing chunk through SSM

        state: (B, H, N)
        u: (B, H, L)

        Returns: (B, H, N)
        """

        if hasattr(self.kernel, "forward_state"):
            return self.kernel.forward_state(u, state)

        dA, dB = self.kernel._setup_state()  # Construct dA, dB matrices
        # dA, dB = self.kernel.dA, self.kernel.dB # (H N N) (H N)

        conj = state.size(-1) != dA.size(-1)
        if conj:
            state = _conj(state)

        v = contract(
            "h n, b h l -> b h n l", dB, u.flip(-1)
        )  # dB.unsqueeze(-1) * u.flip(-1).unsqueeze(-2)
        AL, v = power(u.size(-1), dA, v)
        next_state = contract("h m n, b h n -> b h m", AL, state)
        next_state = next_state + v

        if conj:
            next_state = next_state[..., : next_state.size(-1) // 2]
        return next_state

    def _setup_step(self, **kwargs):
        # This method is intended to be private so that setting up an S4 module with
        # ```
        # if hasattr(module, 'setup_step'): module.setup_step()
        # ```
        # will not trigger this method multiple times
        self.kernel._setup_step(**kwargs)

    def step(self, u, state, **kwargs):
        y, state = self.kernel.step(u, state, **kwargs)
        return y, state

    def default_state(self, *args, **kwargs):
        return self.kernel.default_state(*args, **kwargs)


class S4(nn.Module):
    def __init__(
        self,
        d_model,
        d_state=64,
        l_max=None,
        channels=1,
        bidirectional=False,
        # Arguments for position-wise feedforward components
        activation="gelu",
        postact="glu",
        hyper_act=None,
        dropout=0.0,
        tie_dropout=False,
        bottleneck=None,
        gate=None,
        transposed=True,
        verbose=False,
        # SSM Kernel arguments
        **kernel_args,
    ):
        """
        d_state: the dimension of the state, also denoted by N
        l_max: the maximum kernel length, also denoted by L. Set l_max=None to always use a global kernel
        channels: can be interpreted as a number of "heads"; the SSM is a map from a 1-dim to C-dim sequence. It's not recommended to change this unless desperate for things to tune; instead, increase d_model for larger models
        bidirectional: if True, convolution kernel will be two-sided

        Position-wise feedforward components:
        --------------------
        activation: activation in between SS and FF
        postact: activation after FF
        hyper_act: use a "hypernetwork" multiplication (experimental)
        dropout: standard dropout argument. tie_dropout=True ties the dropout mask across the sequence length, emulating nn.Dropout1d

        Other arguments:
        --------------------
        transposed: choose backbone axis ordering of (B, L, H) (if False) or (B, H, L) (if True) [B=batch size, L=sequence length, H=hidden dimension]
        gate: add gated activation (GSS)
        bottleneck: reduce SSM dimension (GSS)

        See the class SSKernel for the kernel constructor which accepts kernel_args. Relevant options that are worth considering and tuning include "mode" + "measure", "dt_min", "dt_max", "lr"

        Other options are all experimental and should not need to be configured
        """

        super().__init__()
        if verbose:
            log.info(f"Constructing S4 (H, N, L) = ({d_model}, {d_state}, {l_max})")

        self.d_model = d_model
        self.H = d_model
        self.N = d_state
        self.L = l_max
        self.bidirectional = bidirectional
        self.channels = channels
        self.transposed = transposed

        self.gate = gate
        self.bottleneck = bottleneck

        if bottleneck is not None:
            self.H = self.H // bottleneck
            self.input_linear = LinearActivation(
                self.d_model,
                self.H,
                transposed=self.transposed,
                activation=activation,
                activate=True,
            )

        if gate is not None:
            self.input_gate = LinearActivation(
                self.d_model,
                self.d_model * gate,
                transposed=self.transposed,
                activation=activation,
                activate=True,
            )
            self.output_gate = LinearActivation(
                self.d_model * gate,
                self.d_model,
                transposed=self.transposed,
                activation=None,
                activate=False,
            )

        # optional multiplicative modulation GLU-style
        # https://arxiv.org/abs/2002.05202
        self.hyper = hyper_act is not None
        if self.hyper:
            channels *= 2
            self.hyper_activation = Activation(hyper_act)

        self.D = nn.Parameter(torch.randn(channels, self.H))

        if self.bidirectional:
            channels *= 2

        # SSM Kernel
        self.kernel = SSKernel(
            self.H,
            N=self.N,
            L=self.L,
            channels=channels,
            verbose=verbose,
            **kernel_args,
        )

        # Pointwise
        self.activation = Activation(activation)
        dropout_fn = DropoutNd if tie_dropout else nn.Dropout
        self.dropout = dropout_fn(dropout) if dropout > 0.0 else nn.Identity()
        # position-wise output transform to mix features
        self.output_linear = LinearActivation(
            self.H * self.channels,
            self.d_model * (1 if self.gate is None else self.gate),
            transposed=self.transposed,
            activation=postact,
            activate=True,
        )

    def forward(self, u, state=None, rate=1.0, lengths=None, **kwargs):
        """
        u: (B H L) if self.transposed else (B L H)
        state: (H N) never needed unless you know what you're doing

        Returns: same shape as u
        """
        if not self.transposed:
            u = u.transpose(-1, -2)
        L = u.size(-1)

        # Mask out padding tokens
        if isinstance(lengths, int):
            if lengths != L:
                lengths = torch.tensor(lengths, dtype=torch.long, device=u.device)
            else:
                lengths = None
        if lengths is not None:
            assert (
                isinstance(lengths, torch.Tensor)
                and lengths.ndim == 1
                and lengths.size(0) in [1, u.size(0)]
            )
            mask = torch.where(
                torch.arange(L, device=lengths.device) < lengths[:, None, None],
                1.0,
                0.0,
            )
            u = u * mask

        if self.gate is not None:
            v = self.input_gate(u)
        if self.bottleneck is not None:
            u = self.input_linear(u)

        # Compute SS Kernel
        L_kernel = L if self.L is None else min(L, round(self.L / rate))
        k, k_state = self.kernel(
            L=L_kernel, rate=rate, state=state
        )  # (C H L) (B C H L)

        # Convolution
        if self.bidirectional:
            k0, k1 = rearrange(k, "(s c) h l -> s c h l", s=2)
            k = F.pad(k0, (0, L)) + F.pad(k1.flip(-1), (L, 0))
        k_f = torch.fft.rfft(k, n=L_kernel + L)  # (C H L)
        u_f = torch.fft.rfft(u, n=L_kernel + L)  # (B H L)
        y_f = contract("bhl,chl->bchl", u_f, k_f)
        y = torch.fft.irfft(y_f, n=L_kernel + L)[..., :L]  # (B C H L)

        # Compute D term in state space equation - essentially a skip connection
        y = y + contract("bhl,ch->bchl", u, self.D)

        # Compute state update
        if state is not None:
            assert (
                not self.bidirectional
            ), "Bidirectional not supported with state forwarding"
            y = y + k_state  #
            next_state = self.kernel.forward_state(u, state)
        else:
            next_state = None

        # Optional hyper-network multiplication
        if self.hyper:
            y, yh = rearrange(y, "b (s c) h l -> s b c h l", s=2)
            y = self.hyper_activation(yh) * y

        # Reshape to flatten channels
        y = rearrange(y, "... c h l -> ... (c h) l")

        y = self.dropout(self.activation(y))

        if not self.transposed:
            y = y.transpose(-1, -2)

        y = self.output_linear(y)

        if self.gate is not None:
            y = self.output_gate(y * v)

        return y, next_state

    def setup_step(self, **kwargs):
        self.kernel._setup_step(**kwargs)

    def step(self, u, state):
        """Step one time step as a recurrent model. Intended to be used during validation.

        u: (B H)
        state: (B H N)
        Returns: output (B H), state (B H N)
        """
        assert not self.training

        y, next_state = self.kernel.step(u, state)  # (B C H)
        y = y + u.unsqueeze(-2) * self.D
        y = rearrange(y, "b c h -> b (c h)")
        y = self.activation(y)
        if self.transposed:
            y = self.output_linear(y.unsqueeze(-1)).squeeze(-1)
        else:
            y = self.output_linear(y)
        return y, next_state

    def default_state(self, *batch_shape, device=None):
        # kernel is not a SequenceModule so it doesn't need to adhere to same interface
        # the kernel will know the device of its own parameters
        return self.kernel.default_state(*batch_shape)

    @property
    def d_output(self):
        return self.d_model



from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.env import POPGymEnv


class Battleship(POPGymEnv):
    """Classic game of Battleship, except the board is obscured. Instead, the agent
    receives a hit/miss notification with the tile coordinates.

    Args:
        board_size: The width/height of the board in tiles
        ship_sizes: A list of ships sizes (in tiles) to place on the board

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, board_size=10, ship_sizes=[2, 3, 3, 4]):
        # Params
        self.board_size = board_size
        self.ship_sizes = ship_sizes
        self.max_episode_length = self.board_size**2
        self.observation_space = gym.spaces.Discrete(2)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete(
                    np.full((self.board_size, self.board_size), 2, dtype=int)
                ),
                gym.spaces.MultiDiscrete(
                    np.full((self.board_size, self.board_size), 2, dtype=int)
                ),
                self.observation_space,
            )
        )
        self.action_space = gym.spaces.MultiDiscrete([self.board_size, self.board_size])
        self.last_obs = 0

    def get_state(self):
        state = (self.board.copy(), self.guesses.copy(), self.last_obs)
        return state

    def step(self, action):
        hit = False
        is_ship = self.board[action[0], action[1]]
        guessed_before = self.guesses[action[0], action[1]]
        self.guesses[action[0], action[1]] = 1
        if is_ship and not guessed_before:
            hit = True
            self.hits += 1

        self.num_steps += 1

        # Episode finishes when the max steps is reached
        # (which defaults to the number of squares on the board)
        #       or when all of the ships have been sunk
        done = (self.num_steps >= self.max_episode_length) or (
            self.hits == self.needed_hits
        )
        # Obs is 1 for a hit and 0 for a miss
        obs = int(hit)
        # Reward is structured so that an episode of all hits will have a
        # total reward of 1.0,
        #
        # an episode with all misses will have a total reward of
        # -1.0 * (max_steps / (max_steps-num_hits_needed)),
        #
        # and an episode which simply guesses every single square on the
        # board will have a reward of 0.0
        reward = (int(hit) * (1.0 / self.needed_hits)) + (
            int(not hit) * (-1.0 / (self.max_episode_length - self.needed_hits))
        )
        info = {}
        self.last_obs = obs

        return obs, reward, done, info

    def place_ship(self, board, ship_size):
        valid = False
        start = None
        end = None
        while not valid:
            idx_start = self.np_random.integers(self.board_size, size=2)
            direction = self.np_random.integers(2, size=2)
            idx_end = idx_start.copy()
            idx_end[direction[0]] = idx_start[direction[0]] + (-1) ** direction[1] * (
                ship_size - 1
            )
            if np.any(idx_end < 0) or np.any(idx_end >= self.board_size):
                continue
            start = np.minimum(idx_start, idx_end)
            end = np.maximum(idx_start, idx_end)
            valid = np.sum(board[start[0] : end[0] + 1, start[1] : end[1] + 1]) == 0
        board[start[0] : end[0] + 1, start[1] : end[1] + 1] = 1

        return board

    def render(self, mode=None):
        viz = np.chararray(self.board.shape, unicode=True)
        viz[self.board == 0] = "·"
        viz[self.board == 1] = "☐"
        viz[(self.guesses == 1) * (self.board == 1)] = "☒"
        viz[(self.guesses == 1) * (self.board == 0)] = "⚬"

        print(" " + str(viz).replace("[", "").replace("]", "").replace("'", ""))

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:

        super(Battleship, self).reset(seed=seed)
        self.num_steps = 0

        self.board = np.zeros((self.board_size, self.board_size), dtype=np.int8)
        for ship_size in self.ship_sizes:
            self.board = self.place_ship(self.board, ship_size)

        self.guesses = np.zeros((self.board_size, self.board_size), dtype=np.int8)

        self.hits = 0
        obs = 0
        self.last_obs = obs

        self.needed_hits = sum(self.ship_sizes)
        if return_info:
            return obs, {}

        return obs


class BattleshipEasy(Battleship):
    def __init__(self):
        super().__init__(board_size=8, ship_sizes=[2, 3, 3, 4])


class BattleshipMedium(Battleship):
    def __init__(self):
        super().__init__(board_size=10, ship_sizes=[2, 3, 3, 4])


class BattleshipHard(Battleship):
    def __init__(self):
        super().__init__(board_size=12, ship_sizes=[2, 3, 3, 4])



# This script will run the PPO experiments listed in the paper
# Feel free to set the environment variables to whatever envs/models
# you would like to test

import os
from typing import Any, List

import ray
import torch
from ray.tune.registry import register_env  # noqa: F401

import popgym  # noqa: F401
from popgym import wrappers
from popgym.baselines.ray_models.ray_diffnc import DiffNC  # noqa: F401
from popgym.baselines.ray_models.ray_elman import Elman
from popgym.baselines.ray_models.ray_frameconv import Frameconv
from popgym.baselines.ray_models.ray_framestack import Framestack
from popgym.baselines.ray_models.ray_fwp import FastWeightProgrammer
from popgym.baselines.ray_models.ray_gru import GRU
from popgym.baselines.ray_models.ray_indrnn import IndRNN
from popgym.baselines.ray_models.ray_linear_attention import LinearAttention
from popgym.baselines.ray_models.ray_lmu import LMU
from popgym.baselines.ray_models.ray_lstm import LSTM
from popgym.baselines.ray_models.ray_mlp import MLP, BasicMLP
from popgym.baselines.ray_models.ray_s4d import S4D
from popgym.core.env import POPGymEnv


def main():
    env_names: List[Any] = []

    env_types = os.environ.get("POPGYM_EXPERIMENT", "ALL_ENVS")
    desired_models = os.environ.get("POPGYM_MODELS", "ALL")
    num_splits = int(os.environ.get("POPGYM_NUM_SPLITS", 1))
    split_id = int(os.environ.get("POPGYM_SPLIT_ID", 0))
    project_id = os.environ.get("POPGYM_PROJECT", "popgym-debug")
    gpu_per_worker = float(os.environ.get("POPGYM_GPU", 0.25))
    max_steps = int(os.environ.get("POPGYM_STEPS", 10e6))
    storage_path = os.environ.get("POPGYM_STORAGE", "/tmp/ray_results")
    num_samples = int(os.environ.get("POPGYM_SAMPLES", 1))

    # Used for testing
    # Maximum episode length and backprop thru time truncation length
    bptt_cutoff = int(os.environ.get("POPGYM_BPTT_CUTOFF", 1024))
    num_workers = int(os.environ.get("POPGYM_WORKERS", 4))
    num_minibatch = int(os.environ.get("POPGYM_MINIBATCH", 8))
    num_envs_per_worker = int(os.environ.get("POPGYM_ENVS_PER_WORKER", 16))

    # Hidden size of linear layers
    h = 128
    # Hidden size of memory
    h_memory = 256
    train_batch_size = bptt_cutoff * max(num_workers, 1) * num_envs_per_worker

    def wrap(env: POPGymEnv) -> POPGymEnv:
        return wrappers.Antialias(wrappers.PreviousAction(env))

    # Register all envs with ray
    envs = popgym.ALL_ENVS
    for cls, info in envs.items():
        env_name = info["id"]
        register_env(env_name, lambda x: wrap(cls()))

    # Of the registered envs, pick out the ones we actually want to run
    env_names = []
    for e in env_types.split(","):
        # getattr will either return a dict of {class: info[name}}
        # or just a class
        res = getattr(popgym, e)
        if isinstance(res, dict):
            desired_envs = list(res.keys())
        else:
            desired_envs = [res]

        for d in desired_envs:
            env_names.append(envs[d]["id"])

    env_names = env_names[split_id::num_splits]

    # Setup the models we want to train
    attn_models = [
        LinearAttention,
        FastWeightProgrammer,
    ]
    rnn_models = [LSTM, GRU, Elman, LMU, IndRNN, DiffNC]
    conv_models = [S4D]
    basic_models = [
        BasicMLP,
        MLP,
        Framestack,
        Frameconv,
    ]
    models = rnn_models + attn_models + conv_models + basic_models

    # Filter models by env variable
    if desired_models != "ALL":
        models = [m for m in models if m.__name__ in desired_models.split(",")]

    def trial_name(trial):
        env = trial.config["env"].replace("popgym-", "").replace("-v0", "")
        model = trial.config["model"]["custom_model"].__name__
        emb = trial.config["model"]["custom_model_config"].get("embedding", None)
        emb_size = trial.config["model"]["custom_model_config"].get(
            "embedding_size", None
        )

        return "-".join([str(s) for s in [env, model, emb, emb_size] if s is not None])

    config = {
        # Environments or env names
        "env": ray.tune.grid_search(env_names),
        # Should always be torch
        "framework": "torch",
        # Number of rollout workers
        "num_workers": num_workers,
        # Number of envs per rollout worker
        "num_envs_per_worker": num_envs_per_worker,
        # Num gpus used for the train worker
        "num_gpus": gpu_per_worker,
        # Loss coeff for the ppo value function
        "vf_loss_coeff": 1.0,
        # Num transitions in each training epoch
        "train_batch_size": train_batch_size,
        # Chunk size of transitions sent from rollout workers to trainer
        "rollout_fragment_length": bptt_cutoff,
        # Size of minibatches within epoch
        "sgd_minibatch_size": num_minibatch * bptt_cutoff,
        # decay gamma
        "gamma": 0.99,
        # Required due to RLlib PPO bugs
        "horizon": bptt_cutoff,
        # RLlib bug with truncate_episodes:
        # each batch the temporal dim shrinks by one
        # for now, just use complete_episodes
        "batch_mode": "complete_episodes",
        # "min_sample_timesteps_per_reporting": train_batch_size,
        "min_sample_timesteps_per_iteration": train_batch_size,
        # Describe your RL model here
        "model": {
            # Truncate sequences into no more than this many timesteps
            "max_seq_len": bptt_cutoff,
            # Custom model class
            "custom_model": ray.tune.grid_search(models),
            # Config passed to custom model constructor
            # see base_model.py to see how these are used
            "custom_model_config": {
                "preprocessor_input_size": h,
                "preprocessor": torch.nn.Sequential(
                    torch.nn.Linear(h, h),
                    torch.nn.LeakyReLU(inplace=True),
                ),
                "preprocessor_output_size": h,
                "hidden_size": h_memory,
                "postprocessor": torch.nn.Identity(),
                "actor": torch.nn.Sequential(
                    torch.nn.Linear(h_memory, h),
                    torch.nn.LeakyReLU(inplace=True),
                    torch.nn.Linear(h, h),
                    torch.nn.LeakyReLU(inplace=True),
                ),
                "critic": torch.nn.Sequential(
                    torch.nn.Linear(h_memory, h),
                    torch.nn.LeakyReLU(inplace=True),
                    torch.nn.Linear(h, h),
                    torch.nn.LeakyReLU(inplace=True),
                ),
                "postprocessor_output_size": h,
            },
        },
    }
    # When to stop training
    stop = {"timesteps_total": max_steps}

    # Write your own wandb entity here
    if project_id:
        from ray.air.callbacks.wandb import WandbLoggerCallback

        logging_callbacks = [
            WandbLoggerCallback(
                project=project_id, entity="prorok-lab", log_config=True
            )
        ]
    else:
        logging_callbacks = []

    ray.init()
    ray.tune.run(
        "PPO",
        config=config,
        stop=stop,
        callbacks=logging_callbacks,
        trial_name_creator=trial_name,
        verbose=1,
        num_samples=num_samples,
        local_dir=storage_path,
    )


if __name__ == "__main__":
    main()



import multiprocessing
import time

import popgym

NUM_STEPS = 100000
NUM_WORKERS = 2


def run_sample(e, num_steps):
    env = e()
    env.reset()
    start = time.time()
    for i in range(NUM_STEPS):
        _, _, done, _ = env.step(env.action_space.sample())
        if done:
            env.reset()
    end = time.time()
    elapsed = end - start
    fps = NUM_STEPS / elapsed
    return fps


def main():
    for e in popgym.ALL_BASE_ENVS:
        fps = run_sample(e, NUM_STEPS)
        print(f"{e.__name__} (1x) FPS: {fps:.0f}")

    p = multiprocessing.Pool(processes=NUM_WORKERS)
    with p:
        for e in popgym.ALL_ENVS:
            envs = NUM_WORKERS * [e]
            steps = NUM_WORKERS * [int(NUM_STEPS / NUM_WORKERS)]
            fps = sum(p.starmap(run_sample, zip(envs, steps)))
            print(f"{e.__name__} ({NUM_WORKERS}x) FPS: {fps:.0f}")


if __name__ == "__main__":
    main()



import copy
from typing import Callable, List

import gym
import numpy as np


def ascii_version_of_card(ranks, suits, return_string=True):
    """Instead of a boring text version of the card we render an ASCII image of
    the card.

    :param cards: One or more card objects
    :param return_string: By default we return the string version
        of the card, but the dealer hide the 1st card and we
    keep it as a list so that the dealer can add a hidden card in front of the list
    """
    # we will use this to prints the appropriate icons for each card
    suits_name = ["s", "d", "h", "c"]
    suits_symbols = ["♠", "♦", "♥", "♣"]

    # create an empty list of list, each sublist is a line
    lines = [[] for i in range(9)]

    for s, r in zip(suits, ranks):
        # "King" should be "K" and "10" should still be "10"
        if r == "10":  # ten is the only one who's rank is 2 char long
            rank = r
            space = ""  # if we write "10" on the card that line will be 1 char to long
        else:
            rank = r
            space = " "  # no "10", we use a blank space to will the void
        # get the cards suit in two steps
        suit = suits_name.index(s)
        suit = suits_symbols[suit]

        # add the individual card on a line by line basis
        lines[0].append("┌─────────┐")
        lines[1].append(
            "│{}{}       │".format(rank, space)
        )  # use two {} one for char, one for space or char
        lines[2].append("│         │")
        lines[3].append("│         │")
        lines[4].append("│    {}    │".format(suit))
        lines[5].append("│         │")
        lines[6].append("│         │")
        lines[7].append("│       {}{}│".format(space, rank))
        lines[8].append("└─────────┘")

    result = []
    for index, line in enumerate(lines):
        result.append("".join(lines[index]))

    # hidden cards do not use string
    if return_string:
        return "\n".join(result)
    else:
        return result


class DeckEmptyError(Exception):
    pass


RANKS = np.array(["a", "2", "3", "4", "5", "6", "7", "8", "9", "10", "j", "q", "k"])
SUITS = np.array(["s", "d", "c", "h"])
SUITS_UNICODE = ["♠", "♦", "♥", "♣"]
COLORS = np.array(["b", "r"])
DECK_SIZE = 52


class Deck:
    """An object that represents a collection of cards.

    A deck can represent a single deck or multiple decks
    """

    def get_obs_space(self, fields=["colors", "suits", "ranks"], hand_size=1):
        space = []
        for f in fields:
            if f == "colors":
                space.append(COLORS.size)
            elif f == "suits":
                space.append(SUITS.size)
            elif f == "ranks":
                space.append(RANKS.size)
            else:
                raise Exception(f"Invalid field: {f}")
        space = np.tile(np.array(space), hand_size)
        if len(space) == 1:
            return gym.spaces.Discrete(space)
        return gym.spaces.MultiDiscrete(space)

    def __init__(self, num_decks=1, shuffle=False):
        self.num_decks = num_decks
        self.num_cards = DECK_SIZE * num_decks
        self.idx = np.arange(self.num_cards)

        self.ranks = np.tile(RANKS.repeat(SUITS.size), num_decks)
        self.ranks_idx = np.tile(np.arange(RANKS.size).repeat(SUITS.size), num_decks)

        self.suits = np.tile(np.tile(SUITS, RANKS.size), num_decks)
        self.suits_idx = np.tile(np.tile(np.arange(SUITS.size), RANKS.size), num_decks)

        self.colors = np.tile(COLORS, self.num_cards // 2)
        self.colors_idx = np.tile(np.arange(COLORS.size), self.num_cards // 2)
        self.hands = {}
        # The length of the deck, which decreases as cards
        # are drawn/dealt
        self.deck_len = self.num_cards
        if shuffle:
            # For determinism, we should always shuffle from the original input
            self.idx = np.arange(self.num_cards)
            np.random.shuffle(self.idx)

        assert self.idx.size == self.num_cards
        assert self.ranks.size == self.num_cards
        assert self.suits.size == self.num_cards
        assert self.colors.size == self.num_cards
        assert self.ranks_idx.size == self.num_cards
        assert self.suits_idx.size == self.num_cards
        assert self.colors_idx.size == self.num_cards

        self.keys = ["idx", "ranks", "suits", "colors"]
        self.idx_keys = ["ranks_idx", "suits_idx", "colors_idx", "idx"]

    def define_hand_value(
        self, fn: Callable[[List[str]], int], fields: List[str]
    ) -> None:
        """Pass in a function to be used to define the value of a hand
        or set of cards"""
        self.value_fn = fn
        for f in fields:
            assert f in self.keys, f"{f} is not {self.keys}"
        self.value_fn_args = fields

    def clone(self) -> "Deck":
        return copy.deepcopy(self)

    def value(self, player: str) -> int:
        """Returns the value of a players hand by calling the function passed
        to define_hand_value"""
        assert hasattr(
            self, "value_fn"
        ), "Must specify a value fn using define_hand_value first!"

        args = []
        for arg in self.value_fn_args:
            idx = self.hands[player]
            args.append(getattr(self, arg)[idx])

        # Numpy will attempt to unpack an ndarray
        # if it is the only element in a list
        # when using a starred expression
        # i.e. len([np.array([1,2,3]])) == 3
        if len(args) == 1:
            return self.value_fn(args[0])
        else:
            return self.value_fn(*args)

    def value_idx(self, idx: List[int]) -> int:
        """Returns the value of a selection of cards"""
        assert hasattr(
            self, "value_fn"
        ), "Must specify a value fn using define_hand_value first!"
        args = []
        for arg in self.value_fn_args:
            args.append(getattr(self, arg)[idx])

        return self.value_fn(*args)

    def __len__(self):
        return self.deck_len

    def __getitem__(self, item):
        return self.hands[item]

    def add_players(self, *players: List[str]) -> None:
        for player in players:
            self.hands[player] = []

    def deal(self, player: str, num_cards: int = 1) -> None:
        """Deals a number of cards to the specified player
        from the deck"""
        new_len = self.deck_len - num_cards
        if new_len < 0:
            raise DeckEmptyError()

        self.hands[player] += self.idx[new_len : self.deck_len].tolist()
        self.deck_len = new_len

    def discard_hands(self, *players: List[str]):
        """Discards the cards in the hand of a player. Note that
        these cards do not go back into the deck. Call reset()
        to fold the hands back into the deck"""
        for player in players:
            self.hands[player].clear()

    def discard_all(self):
        """Discards the cards in all player hands. Note that
        these cards do not go back into the deck. Call reset()
        to fold the hands back into the deck"""
        for player in self.hands:
            self.hands[player].clear()

    def discard(self, player: str, hand_idx: int):
        """Discards one card in the players hand at the specified idx.
        Note this idx refers to the idx of the card in the hand, rather
        than the idx of the card in the deck"""
        self.hands[player].pop(hand_idx)

    def reset(self, shuffle=True, rng=None):
        """Empties the hands of all players and places cards
        back into the deck in their original position. Optionally
        shuffles the deck afterwards"""
        for hands in self.hands.values():
            hands.clear()
        self.deck_len = self.num_cards
        if shuffle:
            self.idx = np.arange(self.num_cards)
            if rng is None:
                np.random.shuffle(self.idx)
            else:
                self.rng = rng
                rng.shuffle(self.idx)

    def show(
        self, player: str, fields: List[str] = ["colors", "suits", "ranks"], pad_to=None
    ) -> List[np.ndarray]:
        """Shows the hand of the player, returning the fields specified of the cards
        they hold. Optionally zero-pad to a size."""
        reprs = []
        if pad_to is not None:
            padding = [0] * (pad_to - len(self.hands[player]))
        else:
            padding = []
        hand_idx = np.array(self.hands[player] + padding, dtype=np.int64)
        for f in fields:
            assert f in [*self.idx_keys, *self.keys], f"{f} is not a valid key"
            arr = getattr(self, f)
            # Special case, do not double index indices
            if f == "idx":
                reprs.append(np.array(hand_idx))
            else:
                # Requires indexing
                reprs.append(arr[hand_idx])

        return np.stack(reprs)

    def hand_size(self, player: str) -> int:
        return len(self.hands[player])

    def visualize(self, player: str) -> str:
        """Returns a string visualization of a player's hand, for printing
        to the terminal"""
        if len(self[player]) == 0:
            return "\n".join([""] * 10)
        ranks, suits = self.show(player, ["ranks", "suits"])
        return ascii_version_of_card(ranks, suits)

    def visualize_idx(self, idx: List[int]) -> str:
        """Returns a string visualization of the following idx,
        referring to cards in the hand or deck"""
        if len(idx) == 0:
            return "\n".join([""] * 10)
        suits = self.suits[idx]
        ranks = self.ranks[idx]
        return ascii_version_of_card(ranks, suits)



import unittest

import numpy as np

from popgym.envs.labyrinth_explore import LabyrinthExplore


class TestLabyrinthExplore(unittest.TestCase):
    def test_step(self):
        e = LabyrinthExplore()
        obs = e.reset()
        self.assertEqual(obs.dtype, np.int32)
        done = False
        for i in range(1000):
            obs, _, done, _ = e.step(e.action_space.sample())
            self.assertEqual(obs.dtype, np.int32)
            if done:
                obs = e.reset()
                self.assertEqual(obs.dtype, np.int32)

    def test_noclip(self):
        e = LabyrinthExplore()
        done = False
        e.reset()
        for i in range(5):
            while not done:
                _, _, done, _ = e.step(e.action_space.sample())

            visit_mask = e.explored == 1
            obstacle_mask = e.maze.grid == 1
            self.assertFalse(np.any(visit_mask * obstacle_mask == 1))
            e.reset()

    def test_known(self):
        e = LabyrinthExplore((6, 6))
        e.reset(seed=0)
        left, right, up, down = 0, 1, 2, 3
        actions = (
            [left] * 5
            + [up] * 2
            + [right] * 2
            + [left] * 2
            + [down] * 2
            + [right] * 4
            + [up] * 4
            + [left] * 4
        )
        done = False
        cum_rew = 0
        for action in actions:
            self.assertFalse(done)
            obs, reward, done, info = e.step(action)
            cum_rew += reward

        self.assertTrue(done)
        expected = 1.0 + 8 * e.neg_reward_scale
        self.assertTrue(done)
        self.assertAlmostEqual(cum_rew, expected)



import math
from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class Concentration(POPGymEnv):
    """Classic game of concentration. A deck of cards is shuffled and placed
    face-down. The player can flip two cards, if they match they get a reward
    otherwise they dont.

    Args:
        num_decks: Number of decks to play with
        deck_type: String denoting what we are matching. Can be the card colors
        (colors) or the card ranks (ranks)

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, num_decks=1, deck_type="ranks"):
        # See https://math.stackexchange.com/questions/1876467/
        # how-many-turns-on-average-does-it-take-for-a-perfect-player
        # -to-win-concentrati
        self.deck = Deck(num_decks=num_decks)
        n = 52 * num_decks
        self.episode_length = math.ceil(2 * n - (n / (2 * n - 1)))
        self.success_reward_scale = 1 / (self.deck.num_cards // 2)
        self.failure_reward_scale = -1 / (self.episode_length)

        if deck_type == "colors":
            self.deck_type = self.deck.colors
            self.deck_idx_type = self.deck.colors_idx
        # Cannot do suits because there are 13 of each suit -- we need
        # an even number to match
        elif deck_type == "ranks":
            self.deck_type = self.deck.ranks
            self.deck_idx_type = self.deck.ranks_idx
        else:
            raise NotImplementedError(f"Invalid deck type {deck_type}")

        # cards = 14 * np.ones(len(self.deck))
        self.facedown_card = len(np.unique(self.deck_type))
        cards = (1 + self.facedown_card) * np.ones(self.deck.num_cards)
        self.observation_space = gym.spaces.MultiDiscrete(cards)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete(cards - 1),  # cards values
                gym.spaces.MultiBinary(self.deck.num_cards),  # cards faces on
                gym.spaces.Discrete(n),  # first card turned
                gym.spaces.Discrete(n),  # second card turned
            )
        )
        self.action_space = gym.spaces.Discrete(np.array(self.deck.num_cards))
        self.deck.add_players("face_up", "face_up_idx", "in_play", "in_play_idx")
        self.last_in_play_idx = []
        self.last_trying_card_already_up = False

    def step(self, action):
        reward = 0
        done = False

        # Done conditions
        if self.curr_step >= self.episode_length - 1:
            done = True

        # Flip card
        self.deck["in_play"].append(self.deck.idx[action])
        self.deck["in_play_idx"].append(action)

        assert len(self.deck["in_play"]) <= 2
        # Cannot flip the same card twice
        self.obs = self.get_obs()
        self.last_in_play_idx = self.deck["in_play_idx"].copy()

        assert len(self.deck["face_up"]) == len(self.deck["face_up_idx"])
        assert len(self.deck["in_play"]) == len(self.deck["in_play_idx"])

        trying_card_already_up = any(
            idx in self.deck["face_up_idx"] for idx in self.deck["in_play_idx"]
        )

        # End of phase
        if trying_card_already_up:
            reward = self.failure_reward_scale * len(self.deck["in_play_idx"])
            self.deck.discard_hands("in_play", "in_play_idx")
        elif len(self.deck["in_play"]) == 2:
            inplay_cards = self.deck_type[self.deck["in_play"]]
            cards_match = inplay_cards[0] == inplay_cards[1]
            flipped_same_idx = self.deck["in_play"][0] == self.deck["in_play"][1]
            if cards_match and not flipped_same_idx:
                reward = self.success_reward_scale
                self.deck["face_up"].extend(self.deck["in_play"])
                self.deck["face_up_idx"].extend(self.deck["in_play_idx"])
                done |= len(self.deck["face_up"]) == len(self.deck)
            else:
                reward = 2 * self.failure_reward_scale

            # Flip two last flipped-up cards face down again
            self.deck.discard_hands("in_play", "in_play_idx")
            self.hand = []

        self.curr_step += 1

        return self.obs, reward, done, {}

    def render(self, mode="ascii"):
        self.obs.tolist()
        visible_mask = self.obs != self.facedown_card
        rend = np.full(self.obs.shape, "?")
        rend[visible_mask] = self.obs[visible_mask]
        rend = rend.reshape(4, 13)
        output = (
            " "
            + str(rend)
            .replace("[", "")
            .replace("]", "")
            .replace(",", "")
            .replace("'", "")
            # .replace("\n", "")
        )
        print(output)

    def get_obs(self):
        obs = self.facedown_card * np.ones(len(self.deck), dtype=np.int64)
        hand_idx = self.deck["face_up_idx"] + self.deck["in_play_idx"]
        deck_idx = self.deck.idx[hand_idx]
        obs[hand_idx] = self.deck_idx_type[deck_idx]
        return obs

    def get_state(self):
        cards_face_up = np.zeros(
            len(
                self.deck,
            ),
            dtype=np.int8,
        )
        cards_face_up[self.deck["face_up_idx"]] = 1
        first_card, second_card = self.facedown_card, self.facedown_card
        if len(self.last_in_play_idx) == 2:
            first_card, second_card = self.last_in_play_idx
        elif len(self.last_in_play_idx) == 1:
            first_card = self.last_in_play_idx[0]
        return self.state.copy(), cards_face_up, first_card, second_card

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        self.flip_next_turn = False
        self.deck.reset(rng=self.np_random)
        self.state = self.deck_idx_type[self.deck.idx].copy().astype(np.float32)
        self.curr_step = 0
        self.turn_counter = 0
        self.flipped_counter = 0
        self.last_in_play_idx = []
        self.last_trying_card_already_up = False
        self.obs = self.get_obs()
        info: Dict[str, Any] = {}
        if return_info:
            return self.obs, info

        return self.obs


class ConcentrationEasy(Concentration):
    def __init__(self):
        super().__init__(num_decks=1, deck_type="colors")


class ConcentrationMedium(Concentration):
    def __init__(self):
        super().__init__(num_decks=2, deck_type="colors")


class ConcentrationHard(Concentration):
    def __init__(self):
        super().__init__(num_decks=1, deck_type="ranks")



import math
import unittest
import random
import numpy as np


def repeat(times):
    # https://stackoverflow.com/a/13606054
    def repeat_helper(f):
        def call_helper(*args):
            for i in range(0, times):
                f(*args)

        return call_helper

    return repeat_helper


def is_close(x, y):
    assert type(x) is type(y)
    if isinstance(x, (list, tuple)):
        close = all(map(is_close, x, y))
    elif isinstance(x, dict):
        close = all([is_close(x[key], y[key]) for key in x.keys()])
    elif isinstance(x, np.ndarray):
        close = np.allclose(x, y)
    elif isinstance(x, float):
        close = math.isclose(x, y)
    else:
        close = x == y
    return close


class AbstractTest(object):
    class POPGymTest(unittest.TestCase):
        def test_init(self):
            self.env.reset()

        def test_get_state(self):
            self.env.reset()
            self.env.get_state()

        def test_step(self):
            self.env.reset()
            done = False
            while not done:
                obs, reward, done, info = self.env.step(self.env.action_space.sample())

        def test_spaces(self):
            obs = self.env.reset()
            state = self.env.get_state()
            self.assertTrue(self.env.observation_space.contains(obs))
            self.assertTrue(self.env.state_space.contains(state))
            done = False
            while not done:
                obs, reward, done, info = self.env.step(self.env.action_space.sample())
                state = self.env.get_state()
                self.assertTrue(self.env.observation_space.contains(obs))
                self.assertTrue(self.env.state_space.contains(state))

        # @repeat(5)
        def test_np_random(self):
            seed = random.randint(0, 1000)
            state_list = []
            obs_list = []
            reward_list = []
            done = False
            obs = self.env.reset(seed=seed)
            obs_list.append(obs)
            state_list.append(self.env.get_state())
            action_list = []
            while not done:
                action = self.env.action_space.sample()
                action_list.append(action)
                obs, rew, done, info = self.env.step(action)
                obs_list.append(obs)
                reward_list.append(rew)
                state_list.append(self.env.get_state())

            obs = self.env.reset(seed=seed)
            state = self.env.get_state()
            i = 0
            self.assertTrue(is_close(obs, obs_list[i]))
            self.assertTrue(is_close(state, state_list[i]))
            done = False
            while not done:
                obs, rew, done, info = self.env.step(action_list[i])
                state = self.env.get_state()
                self.assertTrue(is_close(obs, obs_list[i + 1]))
                self.assertTrue(is_close(state, state_list[i + 1]))
                self.assertEqual(rew, reward_list[i])
                self.assertEqual(done, i + 1 == len(reward_list))
                i += 1



from typing import List, Tuple

import gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class Elman(BaseModel):
    """The Elman Network (RNN) from

    @article{elman_finding_1990,
        title = {Finding structure in time},
        volume = {14},
        number = {2},
        journal = {Cognitive science},
        author = {Elman, Jeffrey L},
        year = {1990},
        note = {Publisher: Wiley Online Library},
        pages = {179--211},
    }
    """

    MODEL_CONFIG = {
        # Number of recurrent hidden layers in encoder/decoder
        "num_recurrent_layers": 1,
        "benchmark": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        if self.cfg["benchmark"]:
            self.core = nn.RNNCell(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
            )
        else:
            self.core = nn.RNN(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
                self.cfg["num_recurrent_layers"],
                batch_first=True,
            )

    def initial_state(self) -> List[TensorType]:
        return [torch.zeros(1, self.cfg["hidden_size"])]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        state = state[0].permute(1, 0, 2)

        if self.cfg["benchmark"]:
            outs = []
            state = state.squeeze(0)  # type: ignore
            for t in range(z.shape[1]):
                state = self.core(z[:, t], state)
                outs.append(state)
            z = torch.stack(outs, dim=1)
            state = state.unsqueeze(0)  # type: ignore
        else:
            z, state = self.core(z, state)

        state = [state.permute(1, 0, 2)]  # type: ignore

        return z, state



from typing import Tuple

from gym import spaces
from gym.core import ActType, ObsType

from popgym.core.observability import OBS, STATE
from popgym.core.wrapper import POPGymWrapper
from popgym.core.env import POPGymEnv

IS_T0 = "is_t0"


class Antialias(POPGymWrapper):
    """Wrapper that undoes aliasing produces by the PreviousAction wrapper

    Outputs a boolean flag denoting whether the observation was taken
    at the first timestep.

    Args:
        env: The environment

    Returns:
        A gym environment with a Discrete(2) appended to the observation space
    """

    def __init__(self, env: POPGymEnv):
        super().__init__(env)
        self.observation_space = Antialias.antialias_obs_space(
            self.env.observation_space
        )

    @staticmethod
    def antialias_obs_space(observation_space: spaces.Space) -> spaces.Space:
        flag_space = spaces.Discrete(2)
        if isinstance(
            observation_space,
            (spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.MultiBinary),
        ):
            observation_space = spaces.Tuple((observation_space, flag_space))
        elif isinstance(observation_space, spaces.Tuple):
            observation_space = spaces.Tuple(
                tuple(observation_space.spaces) + (flag_space,)
            )
        elif isinstance(observation_space, spaces.Dict):
            observation_space = observation_space.spaces.copy()
            if set(observation_space.keys()) == {OBS, STATE}:
                # Observation comes from ObservabilityWrapper
                # with observability level FULL_AND_PARTIAL
                observation_space[OBS] = Antialias.antialias_obs_space(
                    observation_space[OBS]
                )
            else:
                observation_space[IS_T0] = flag_space
            observation_space = spaces.Dict(observation_space)
        else:
            raise NotImplementedError("Unknown observation space")
        return observation_space

    @staticmethod
    def antialias_obs(
        observation_space: spaces.Space, obs: ObsType, is_t0: bool
    ) -> ObsType:
        is_t0_asint = int(is_t0)
        if isinstance(
            observation_space,
            (spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.MultiBinary),
        ):
            obs = (obs, is_t0_asint)
        elif isinstance(observation_space, spaces.Tuple):
            obs = tuple(obs) + (is_t0_asint,)
        elif isinstance(observation_space, spaces.Dict):
            if set(observation_space.keys()) == {OBS, STATE}:
                # Observation comes from ObservabilityWrapper
                # with observability level FULL_AND_PARTIAL
                obs[OBS] = Antialias.antialias_obs(
                    observation_space[OBS], obs[OBS], is_t0
                )
            else:
                obs[IS_T0] = is_t0_asint
        else:
            raise NotImplementedError("Unknown observation space")
        return obs

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, dict]:
        obs, reward, done, info = self.env.step(action)
        obs = Antialias.antialias_obs(self.env.observation_space, obs, False)
        return obs, reward, done, info

    def reset(self, **kwargs):
        if kwargs.get("return_info", False):
            obs, info = self.env.reset(**kwargs)
            obs = self.antialias_obs(self.env.observation_space, obs, True)
            return obs, info
        else:
            obs = self.env.reset(**kwargs)
            obs = self.antialias_obs(self.env.observation_space, obs, True)
            return obs



import unittest

import numpy as np

from popgym.envs.multiarmed_bandit import MultiarmedBandit
from tests.base_env_test import AbstractTest


class TestMultiarmedBandit(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = MultiarmedBandit()

    def test_best(self):
        obs, info = self.env.reset(return_info=True)
        bandits = info["bandits"]
        action = np.argmax(bandits)
        done = False
        reward = 0
        expected_value = (
                                 self.env.max_episode_length * np.max(bandits)
                                 - self.env.max_episode_length * (1 - np.max(bandits))
        ) / self.env.max_episode_length

        for i in range(self.env.max_episode_length):
            self.assertFalse(done)
            obs, rew, done, info = self.env.step(action)
            reward += rew

        self.assertTrue(done)
        self.assertTrue(np.abs(expected_value - reward) < 0.1)



import enum
from typing import Any, Dict, Optional, Tuple, Union

import gym
import numpy as np

from popgym.core.deck import RANKS, Deck, DeckEmptyError


class Phase(enum.IntEnum):
    # The betting phase where the player selects a bet
    BET = 0
    # The cards are dealt to the player and house, player does not act
    DEAL = 1
    # Player chooses to hit or stay
    PLAY = 2
    # Player receives the final cards (for counting) and the reward
    PAYOUT = 3


def hand_value(hand):
    card_map = {
        "a": 1,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
        "9": 9,
        "10": 10,
        "j": 10,
        "q": 10,
        "k": 10,
    }
    value = 0
    has_ace = False
    for rank in hand:
        if rank == "a":
            has_ace = True
        value += card_map[rank]

    if value < 11 and has_ace:
        # ace = 1 + 10
        value += 10
    return value


class BlackJack(gym.Env):
    """A game of blackjack, where card counting is possible. Successful agents
    should learn to count cards, and bet higher/hit less often when the deck
    contains higher cards. Note that splitting is not allowed.

    Args:
        bet_sizes: The bet sizes available to the agent. These correspond to
            the final reward.
        num_decks: The number of individual decks combined into a single deck.
            In Vegas, this is usually between four and eight.
        max_rounds: The maximum number of rounds where the agent and dealer
            can hit/stay. There is no max in real blackjack, however this
            would result in a very large observation space.
        games_per_episode: The number of games per episode. This must be set high
            for card-counting to have an effect. When set to one, the game
            becomes fully observable.

    Returns:
        A gym environment
    """

    def __init__(
        self,
        bet_sizes=[0, 1 / 80, 1 / 40, 1 / 20],
        num_decks=1,
        max_rounds=6,
        games_per_episode=20,
    ):
        self.bet_sizes = bet_sizes
        self.max_rounds = max_rounds

        self.deck = Deck(num_decks=num_decks)
        self.deck.define_hand_value(hand_value, ["ranks"])
        self.deck.add_players("dealer", "player")

        # Hit, stay, and bet amount
        self.action_space = gym.spaces.Dict(
            {
                "hit": gym.spaces.Discrete(2),
                "bet_size": gym.spaces.Discrete(len(bet_sizes)),
            }
        )
        self.games_per_episode = games_per_episode
        self.curr_game = 0
        self.curr_round = 0
        self.observation_space = gym.spaces.Dict(
            {
                "phase": gym.spaces.Discrete(3),
                "dealer_hand": gym.spaces.MultiDiscrete(max_rounds * [RANKS.size]),
                "dealer_hand_cards_in_play": gym.spaces.MultiBinary(max_rounds),
                "player_hand": gym.spaces.MultiDiscrete(max_rounds * [RANKS.size]),
                "player_hand_cards_in_play": gym.spaces.MultiBinary(max_rounds),
            }
        )
        self.action_phase = Phase.BET

    def bet(self, action):
        # In bet action mode, we don't do anything
        # we inform the player to place a bet
        # which we set during the deal phase
        result = f"placed bet of {self.curr_bet}"
        return result

    def deal(self, action):
        # First round, serve the cards
        # TODO: Do not take bet twice...
        self.curr_bet = self.bet_sizes[action["bet_size"]]
        self.deck.deal("player", 2)
        self.deck.deal("dealer", 1)

    def play(self, action):
        player_hit = action["hit"]
        if player_hit:
            # Hit
            self.deck.deal("player", 1)
            result = "player hits"

        player_value = self.deck.value("player")
        player_bust = player_value > 21
        player_blackjack = player_value == 21
        player_natural = player_blackjack and self.deck.hand_size("player") == 2
        player_max_cards = self.deck.hand_size("player") == self.max_rounds

        game_done = (
            not player_hit or player_bust or player_blackjack or player_max_cards
        )
        if not game_done:
            return 0, game_done, result

        dealer_plays = (not player_bust and not player_hit) or player_blackjack

        if dealer_plays:
            dealer_value = self.deck.value("dealer")
            while dealer_value < 17 and self.deck.hand_size("dealer") < self.max_rounds:
                self.deck.deal("dealer", 1)
                dealer_value = self.deck.value("dealer")
        else:
            dealer_value = self.deck.value("dealer")

        dealer_bust = dealer_value > 21
        dealer_blackjack = dealer_value == 21
        dealer_natural = dealer_blackjack and self.deck.hand_size("dealer") == 2
        player_adv = player_value - dealer_value

        # compare
        if player_adv == 0:
            reward = 0
            result = f"player ({player_value}) and dealer ({dealer_value}) push"
        elif player_adv > 0:
            reward = self.curr_bet
            result = f"player ({player_value}) beats dealer ({dealer_value})"
        elif player_adv < 0:
            reward = -self.curr_bet
            result = f"player ({player_value}) loses to dealer ({dealer_value})"

        # busts
        if player_bust and not dealer_bust:
            reward = -self.curr_bet
            result = f"player ({player_value}) bust"
        elif dealer_bust and not player_bust:
            reward = self.curr_bet
            result = f"dealer ({dealer_value}) bust"
        elif dealer_bust and player_bust:
            reward = 0
            result = f"player ({player_value}) and dealer ({dealer_value}) bust"

        # naturals
        if player_natural and not dealer_natural:
            reward = 1.5 * self.curr_bet
            result = "player natural"
        elif dealer_natural and not player_natural:
            reward = -self.curr_bet
            result = "dealer natural"
        elif dealer_natural and player_natural:
            result = "push: player and dealer naturals"
            reward = 0

        return reward, game_done, result

    def game_reset(self):
        """Resets a game, but not the entire env."""
        self.curr_game += 1
        self.curr_round = 0
        self.deck.discard_all()

    def step(self, action):
        reward = 0
        done = False
        result = ""

        if len(self.deck) < 3:
            done = True
            result = "deck empty, episode over"

        try:
            game_done = False
            if self.action_phase == Phase.BET:
                result = self.bet(action)
                self.obs, self.info = self.build_obs_infos(result)
                self.action_phase = Phase.DEAL
            elif self.action_phase == Phase.DEAL or self.action_phase == Phase.PAYOUT:
                self.deal(action)
                self.obs, self.info = self.build_obs_infos(result)
                self.action_phase = Phase.PLAY
            elif self.action_phase == Phase.PLAY:
                reward, game_done, result = self.play(action)
                if game_done:
                    self.action_phase = Phase.PAYOUT
                if reward != 0:
                    assert self.action_phase == Phase.PAYOUT
                self.obs, self.info = self.build_obs_infos(result)
        except DeckEmptyError:
            done = True
            self.info["result"] += ", deck empty, episode over"

        self.curr_round += 1

        if game_done:
            # Game done, reset and goes to bet phase
            self.game_reset()
            self.action_phase = Phase.BET

        if self.curr_game == self.games_per_episode - 1:
            done = True

        return self.obs, reward, done, self.info

    def render(self):
        phase = Phase(self.obs["phase"]).name
        print(f"Phase: {phase}")
        print(f"Current Bet: {self.info['current_bet']}")
        dealer = self.deck.visualize_idx(self.info["dealer_hand_idx"])
        player = self.deck.visualize_idx(self.info["player_hand_idx"])
        dealer_val = self.deck.value_idx(self.info["dealer_hand_idx"])
        player_val = self.deck.value_idx(self.info["player_hand_idx"])
        print(f"dealer hand (sum={dealer_val}):\n{dealer}")
        print(f"player hand (sum={player_val}):\n{player}")
        print(self.info["result"])
        print("_______________________________")

    def build_obs_infos(self, result=""):
        # Convert card ids to color, suit, rank
        dealer = self.deck.show("dealer", ["ranks_idx"], pad_to=self.max_rounds)[0]
        dealer_size = self.deck.hand_size("dealer")
        dealer_hand_cards_in_play = np.zeros(self.max_rounds, dtype=np.int8)
        dealer_hand_cards_in_play[:dealer_size] = 1

        # Convert card ids to color, suit, rank
        player = self.deck.show("player", ["ranks_idx"], pad_to=self.max_rounds)[0]
        player_size = self.deck.hand_size("player")
        player_hand_cards_in_play = np.zeros(self.max_rounds, dtype=np.int8)
        player_hand_cards_in_play[:player_size] = 1

        obs: Dict[str, Any] = {
            "phase": self.action_phase.value,
            "dealer_hand": dealer.copy(),
            "dealer_hand_cards_in_play": dealer_hand_cards_in_play,
            "player_hand": player.copy(),
            "player_hand_cards_in_play": player_hand_cards_in_play,
        }
        infos: Dict[str, Any] = {
            "phase": self.action_phase,
            "current_bet": self.curr_bet,
            "dealer_hand_idx": self.deck.show("dealer", ["idx"])[0],
            "dealer_value": self.deck.value("dealer"),
            "player_hand_idx": self.deck.show("player", ["idx"])[0],
            "player_value": self.deck.value("player"),
            "result": result,
        }

        return obs, infos

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, Dict[str, Any]]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.curr_game = 0
        self.curr_round = 0
        self.action_phase = Phase.BET
        self.curr_bet = -float("inf")
        self.obs, self.info = self.build_obs_infos()
        self.action_phase = Phase.DEAL
        if return_info:
            return self.obs, self.info

        return self.obs



import unittest

from popgym.envs.blackjack import BlackJack


class TestBlackjack(unittest.TestCase):
    def test_reset(self):
        b = BlackJack()
        b.reset()

    def test_step(self):
        b = BlackJack()
        b.reset()
        a = {"hit": 1, "bet_size": 1}
        [b.step(a) for i in range(10)]

    def test_many_step(self):
        b = BlackJack()
        b.reset()
        a = {"hit": 1, "bet_size": 1}
        for i in range(2000):
            obs, reward, done, info = b.step(a)
            if done:
                b.reset()

    def test_render(self):
        b = BlackJack()
        b.reset()
        a = {"hit": 1, "bet_size": 1}
        [b.step(a) for i in range(10)]
        b.render()



import unittest

import numpy as np

from popgym.core import deck


class TestCore(unittest.TestCase):
    def test_deck_size(self):
        d = deck.Deck(num_decks=1)
        self.assertEqual(len(d.idx), 52)
        d = deck.Deck(num_decks=7)
        self.assertEqual(len(d.idx), 7 * 52)

    def test_deal_discard_reset(self):
        d = deck.Deck(shuffle=False)
        orig_deck = d.idx.copy()
        d.add_players("a", "b")
        d.deal("a", 3)
        d.deal("b", 2)
        self.assertEqual(d.hand_size("a"), 3)
        self.assertEqual(d.hand_size("b"), 2)
        self.assertEqual(len(d), 52 - 5)

        d.discard_hands("a", "b")
        self.assertEqual(len(d), 52 - 5)

        d.reset(shuffle=False)
        self.assertEqual(len(d), 52)
        self.assertTrue(np.all(orig_deck == d.idx))

    def test_draw_52(self):
        d = deck.Deck()
        orig_deck = d.idx.copy()
        d.add_players("a")
        d.deal("a", 52)
        self.assertTrue(np.all(orig_deck == d.show("a", ["idx"])[0]))

    def test_deal_seq(self):
        d = deck.Deck(shuffle=False)
        d.add_players("a")
        num = 10
        for i in range(num):
            d.deal("a", 1)

        target = d.idx[52 : 52 - num - 1 : -1].tolist()
        self.assertEqual(len(target), num)
        self.assertTrue(d["a"] == target)

    def test_value(self):
        def vf(x):
            cmap = {
                "a": 0,
                "1": 2,
                "2": 4,
                "3": 6,
                "4": 8,
                "5": 10,
                "6": 12,
                "7": 14,
                "8": 16,
                "9": 18,
                "10": 20,
                "j": 21,
                "q": 22,
                "k": 23,
            }
            total = 0
            for c in x:
                total += cmap[c]
            return total

        d = deck.Deck(shuffle=True)
        d.define_hand_value(vf, ["ranks"])
        d.add_players("a")
        d.deal("a", 52)
        value = d.value("a")
        target = vf(d.ranks[d["a"]])
        self.assertEqual(target, value)



import unittest

import numpy as np

from popgym.envs.count_recall import CountRecall
from tests.base_env_test import AbstractTest


class TestCountRecall(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = CountRecall()

    def test_perfect(self):
        counts = {k: 0 for k in range(self.env.num_distinct_cards)}

        obs, info = self.env.reset(return_info=True)
        d, q = obs
        counts[d] += 1
        action = np.array([counts[q]])
        reward = 0
        done = False

        t = 0
        while not done:
            obs, rew, done, info = self.env.step(action)
            d, q = obs
            counts[d] += 1
            action = np.array([counts[q]])
            self.assertEqual(rew, 1 / (self.env.value_deck.num_cards - 1))
            reward += rew
            t += 1

        self.assertEqual(t, self.env.max_episode_length)
        self.assertAlmostEqual(reward, 1.0)

    def test_awful(self):
        counts = {k: 0 for k in range(self.env.num_distinct_cards)}

        obs, info = self.env.reset(return_info=True)
        d, q = obs
        counts[d] += 1
        action = np.array([counts[q] + 2])
        reward = 0
        done = False

        t = 0
        while not done:
            obs, rew, done, info = self.env.step(action)
            d, q = obs
            counts[d] += 1
            action = np.array([counts[q] + 2])
            reward += rew
            t += 1

        self.assertEqual(t, self.env.max_episode_length)
        self.assertLess(reward, 1.0 - 0.1)



# Inspired by ray rllib at
# https://github.com/ray-project/ray/blob/master/rllib/examples/env/stateless_cartpole.py

import random
from typing import Optional, Tuple, Union

import gym
import numpy as np
from gym.spaces import Box

from popgym.envs.stateless_cartpole import StatelessCartPole


class IrregularStatelessCartPole(StatelessCartPole):
    """Partially observable variant of the CartPole gym environment.
    https://github.com/openai/gym/blob/master/gym/envs/classic_control/
    cartpole.py
    We delete the x- and angular velocity components of the state, so that it
    can only be solved by a memory enhanced model (policy).
    """

    def __init__(self, *args, **kwargs):
        original_tau = 0.02
        self.tau_range = kwargs.pop("tau_range", (0.02, 0.03))
        max_diff = max(self.tau_range) / original_tau
        super().__init__(*args, **kwargs)

        # Fix our observation-space (remove 2 velocity components).
        high = np.array(
            [
                self.x_threshold * 2,
                self.theta_threshold_radians * 2,
                max(self.tau_range),
            ],
            dtype=np.float32,
        )

        self.observation_space = Box(
            low=-high * max_diff, high=high * max_diff, dtype=np.float32
        )

    def step(
        self, action: gym.core.ActType
    ) -> Tuple[gym.core.ObsType, float, bool, dict]:
        self.tau = random.uniform(*self.tau_range)
        next_obs, reward, done, info = super().step(action)
        # next_obs is [x-pos, x-veloc, angle, angle-veloc]
        return np.array([*next_obs, self.tau], dtype=np.float32), reward, done, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None
    ) -> Union[gym.core.ObsType, Tuple[gym.core.ObsType, dict]]:
        self.num_steps = 0
        if return_info:
            init_obs, info = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            return np.array([*init_obs, self.tau], dtype=np.float32), info
        else:
            init_obs = super().reset(
                seed=seed, return_info=return_info, options=options
            )
            # init_obs is [x-pos, x-veloc, angle, angle-veloc]
            return np.array([*init_obs, self.tau], dtype=np.float32)


class IrregularStatelessCartPoleEasy(IrregularStatelessCartPole):
    pass


class IrregularStatelessCartPoleMedium(IrregularStatelessCartPole):
    def __init__(self, *args, **kwargs):
        kwargs["tau_range"] = (0.02, 0.04)
        super().__init__(*args, **kwargs)


class IrregularStatelessCartPoleHard(IrregularStatelessCartPole):
    def __init__(self, *args, **kwargs):
        kwargs["tau_range"] = (0.02, 0.05)
        super().__init__(*args, **kwargs)
